# -*- python -*-
# $RCSfile: anisocijkl.py,v $
# $Revision: 1.28 $
# $Author: redwing $
# $Date: 2005/02/02 21:45:23 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov. 

# Anisotropic Cijkl objects are mostly non-registered non-convertible
# structured types, with the possible exception of Cubic.

from common import registeredclass
from common.IO import parameter
from common.IO import xmlmenudump
import struct

FloatParameter = parameter.FloatParameter
Comparable = parameter.Comparable

from common import utils
# Base class.  Should be able to answer questions about its values
# in different representations, using the member registrations'
# to_base and from_base functionality.  This works differently from the
# otherwise-prototypical "Color" class, because the conversion calls
# here are relatively infrequent, so a slightly cleaner structure can
# be used.  This (hopefully) avoids putting conversion code in
# more than one place.
class CubicRank4Tensor(registeredclass.ConvertibleRegisteredClass):
    registry = []

CubicRank4Tensor.tip = "Representations of a cubic 4th rank tensor."
CubicRank4Tensor.discussion = "<para>TODO: Add discussion.</para>"
    
# Base value type for the CubicRank4Tensor.  This is necessary for convertible
# types, because the from_base conversion routine needs to have some
# confidence that it's getting a value it can cope with, which it
# does by type-checking against this type.
class CubicCijklValueBase:
    def __init__(self,c11,c12,c44):
        self.c11 = c11
        self.c12 = c12
        self.c44 = c44
        

# Registration object hosts the "call points" for to_base
# routines.  These are used by the GUI when switching types.
class CubicCijklRegistration(registeredclass.ConvertibleRegistration):
    def __init__(self, name, subclass, ordering,
                 from_base, to_base, params=[], tip=None):
        registeredclass.Registration.__init__(self, name, CubicRank4Tensor,
                                              subclass,
                                              ordering, from_base=from_base,
                                              to_base=to_base, params=params,
                                              tip=tip)
    def getParamValuesAsBase(self):
        return self.to_base(self)

# Define a special Parameter type so that it can have special widgets
# assigned to it in GUI mode.

class CubicRank4TensorParameter(parameter.ConvertibleRegisteredParameter):
    def __init__(self, name, value=None, default=None, tip=None):
        parameter.ConvertibleRegisteredParameter.__init__(
            self, name, CubicRank4Tensor, value=value, default=default, tip=tip)
    def clone(self):
        return CubicRank4TensorParameter(self.name, self.value, self.default,
                                     self.tip)
    def valueDesc(self):
        return "A <link linkend='RegisteredClass:CubicRank4Tensor'><classname>CubicRank4Tensor</classname></link> object."

#########################################################################
#########################################################################

# Conversion for the symmetric matrix representation of the Cijkl
# matrix.  This uses the same internal data representation as
# the base value type, so the to_base and from_base routines
# are conspicuously trivial.

class CubicRank4TensorCij(CubicRank4Tensor):
    def __init__(self, c11, c12, c44):
        # Set the parameters in the registration object.  This is
        # redundant (but harmless) when the object is instanced from
        # the registration.
        self.c11 = c11
        self.c12 = c12
        self.c44 = c44

def _cij_to_base(cijkl_reg, values=None):
    vset = values or cijkl_reg.getParamValues() # This case is trivial.
    return CubicCijklValueBase(c11=vset[0], c12=vset[1], c44=vset[2])

def _cij_from_base(cijkl_reg, base):
    if isinstance(base, CubicCijklValueBase):
        cijkl_reg.setDefaultParams([base.c11,base.c12,base.c44])

CubicCijklRegistration('Cij', CubicRank4TensorCij, 100, from_base=_cij_from_base,
                       to_base=_cij_to_base,
                       params=[FloatParameter('c11', 10.0),
                               FloatParameter('c12', 5.0),
                               FloatParameter('c44', 2.5)] )



# Lame coefficients for cubic elasticity.
class CubicRank4TensorLame(CubicRank4Tensor):
    def __init__(self, lmbda, mu, aniso):
        self.lmbda = lmbda
        self.mu = mu
        self.aniso = aniso

def _lame_to_base(lame_reg, values=None):
    (lmbda, mu, aniso) = values or lame_reg.getParamValues()
    c11 = lmbda + 2.0*mu
    c12 = lmbda
    c44 = 0.5*(c11-c12)*aniso
    return CubicCijklValueBase(c11=c11, c12=c12, c44=c44)

def _lame_from_base(lame_reg, base):
    if isinstance(base, CubicCijklValueBase):
        lmbda = base.c12
        mu = (base.c11-base.c12)/2.0
        aniso = 2.0*base.c44/(base.c11-base.c12)
        lame_reg.setDefaultParams([lmbda,mu,aniso])

CubicCijklRegistration('Lame', CubicRank4TensorLame, 101, from_base=_lame_from_base,
                     to_base=_lame_to_base,
                     params=[FloatParameter('lmbda',10.0),
                             FloatParameter('mu',5.0),
                             FloatParameter('aniso', 1.0)] )


# Young modulus and Poisson ratio.
class CubicRank4TensorEnu(CubicRank4Tensor):
    def __init__(self, young, poisson, aniso):
        self.young = young
        self.poisson = poisson
        self.aniso = aniso

def _enu_to_base(enu_reg, values=None):
    (young, poisson, aniso) = values or enu_reg.getParamValues()
    c11 = young*(1.0-poisson)/((1.0+poisson)*(1.0-2.0*poisson))
    c12 = young*poisson/((1.0+poisson)*(1.0-2.0*poisson))
    c44 = 0.5*(c11-c12)*aniso
    return CubicCijklValueBase(c11=c11, c12=c12, c44=c44)

def _enu_from_base(enu_reg, base):
    if isinstance(base, CubicCijklValueBase):
        c11 = base.c11
        c12 = base.c12
        young = ((c11-c12)*(c11+2.0*c12)/(c11+c12))
        poisson = c12/(c11+c12)
        aniso = 2.0*base.c44/(c11-c12)
        enu_reg.setDefaultParams([young, poisson, aniso])

CubicCijklRegistration('E and nu', CubicRank4TensorEnu, 102,
                       from_base=_enu_from_base,
                       to_base=_enu_to_base,
                       params=[FloatParameter('young',31.6),
                               FloatParameter('poisson',14.7),
                               FloatParameter('aniso',  1.0)] )


# Bulk and shear modulus scheme.
class CubicRank4TensorBulkShear(CubicRank4Tensor):
    def __init__(self, bulk, shear, aniso):
        self.bulk = bulk
        self.shear = shear
        self.aniso = aniso

def _bs_to_base(bs_reg, values=None):
    (bulk, shear, aniso) = values or  bs_reg.getParamValues()
    c11 = bulk+(4.0/3.0)*shear
    c12 = bulk-(2.0/3.0)*shear
    c44 = 0.5*(c11-c12)*aniso
    return CubicCijklValueBase(c11=c11, c12=c12, c44=c44)

def _bs_from_base(bs_reg, base):
    if isinstance(base, CubicCijklValueBase):
        bulk = (1.0/3.0)*(base.c11+2.0*base.c12)
        shear = (base.c11-base.c12)/2.0
        aniso = 2.0*base.c44/(base.c11-base.c12)
        bs_reg.setDefaultParams([bulk, shear, aniso])

CubicCijklRegistration('Bulk and Shear', CubicRank4TensorBulkShear, 103,
                     from_base=_bs_from_base,
                     to_base=_bs_to_base,
                     params=[FloatParameter('bulk',31.6),
                             FloatParameter('shear',14.7),
                             FloatParameter('aniso',1.0)] )


###########################################################################
###########################################################################

# Other crystallographic symmetries do not have multiple representations
# (yet).  They still have custom types, though, so that there can
# be special matrix-like widgets for them, and so that it is an error
# to attempt to assign outside the allowed variables.

# Because these modulus objects are not registered classes, they need
# explicit reprs and need to be explicitly injected into the main OOF
# namespace.

class HexagonalRank4TensorCij(Comparable):
    def __init__(self,c11,c12,c13,c33,c44):
        self.c11 = c11
        self.c12 = c12
        self.c13 = c13
        self.c33 = c33
        self.c44 = c44
    def __repr__(self):
        return "HexagonalRank4TensorCij(c11=%g, c12=%g, c13=%g, c33=%g, c44=%g)" % \
               (self.c11, self.c12, self.c13, self.c33, self.c44)

utils.OOFdefine('HexagonalRank4TensorCij', HexagonalRank4TensorCij)

class HexagonalCijklParameter(parameter.ParameterBase):
    def __init__(self, name, value=None, default=None, tip=None):
        self.checker = parameter.TypeChecker(HexagonalRank4TensorCij)
        parameter.ParameterBase.__init__(self, name,
                                         value=value, default=default,
                                         tip=tip)
    packfmt = ">5d"
    packsize = struct.calcsize(packfmt)
    def binaryRepr(self, datafile, value):
        return struct.pack(HexagonalCijklParameter.packfmt,
                           value.c11, value.c12, value.c13, value.c33,
                           value.c44)
    def binaryRead(self, parser):
        b = parser.getBytes(HexagonalCijklParameter.packsize)
        args = struct.unpack(HexagonalCijklParameter.packfmt, b)
        return HexagonalRank4TensorCij(*args)
    def valueDesc(self):
        return "A <link linkend='Object:HexagonalRank4TensorCij'><classname>HexagonalRank4TensorCij</classname></link> object."

xmlmenudump.XMLObjectDoc(
    'HexagonalRank4TensorCij',
    xmlmenudump.loadFile('DISCUSSIONS/engine/IO/object-anisocijkl.xml'))

class TetragonalRank4TensorCij(Comparable):
    def __init__(self,c11,c12,c13,c33,c44,c66,c16):
        self.c11 = c11
        self.c12 = c12
        self.c13 = c13
        self.c16 = c16
        self.c33 = c33
        self.c44 = c44
        self.c66 = c66
    def __repr__(self):
        return "TetragonalRank4TensorCij(c11=%g, c12=%g, c13=%g, c33=%g, c44=%g, c66=%g, c16=%g)" % \
               (self.c11, self.c12, self.c13, self.c33, self.c44,
                self.c66, self.c16)

utils.OOFdefine('TetragonalRank4TensorCij', TetragonalRank4TensorCij)

# Parameter for this class.
class TetragonalCijklParameter(parameter.ParameterBase):
    def __init__(self, name, value=None, default=None, tip=None):
        self.checker = parameter.TypeChecker(TetragonalRank4TensorCij)
        parameter.ParameterBase.__init__(self, name,
                                         value=value, default=default,
                                         tip=tip)
    packfmt = ">7d"
    packsize = struct.calcsize(packfmt)
    def binaryRepr(self, datafile, value):
        return struct.pack(TetragonalCijklParameter.packfmt,
                           value.c11, value.c12, value.c13, value.c16,
                           value.c33, value.c44, value.c66)
    def binaryRead(self, parser):
        b = parser.getBytes(TetragonalCijklParameter.packsize)
        args = struct.unpack(TetragonalCijklParameter.packfmt, b)
        return TetragonalRank4TensorCij(*args)
    def valueDesc(self):
        return "A <link linkend='Object:TetragonalRank4TensorCij'><classname>TetragonalRank4TensorCij</classname></link> object."

xmlmenudump.XMLObjectDoc(
    'TetragonalRank4TensorCij',
    xmlmenudump.loadFile('DISCUSSIONS/engine/IO/object-tetragonalrank4.xml'))

class TrigonalARank4TensorCij(Comparable):
    def __init__(self,c11,c12,c13,c33,c44,c14,c15):
        self.c11 = c11
        self.c12 = c12
        self.c13 = c13
        self.c33 = c33
        self.c44 = c44
        self.c14 = c14
        self.c15 = c15
    def __repr__(self):
        return "TrigonalARank4TensorCij(c11=%g, c12=%g, c13=%g, c33=%g, c44=%g, c14=%g, c15=%g)" % \
               (self.c11, self.c12, self.c13, self.c33, self.c44,
                self.c14, self.c15)

utils.OOFdefine('TrigonalARank4TensorCij', TrigonalARank4TensorCij)

# Parameter for this class.
class TrigonalACijklParameter(parameter.ParameterBase):
    def __init__(self, name, value=None, default=None, tip=None):
        self.checker = parameter.TypeChecker(TrigonalARank4TensorCij)
        parameter.ParameterBase.__init__(self, name,
                                         value=value, default=default,
                                         tip=tip)
    packfmt = ">7d"
    packsize = struct.calcsize(packfmt)
    def binaryRepr(self, datafile, value):
        return struct.pack(TrigonalACijklParameter.packfmt,
                           value.c11, value.c12, value.c13,
                           value.c33, value.c44, value.c14, value.c15)
    def binaryRead(self, parser):
        b = parser.getBytes(TrigonalACijklParameter.packsize)
        args = struct.unpack(TrigonalACijklParameter.packfmt, b)
        return TrigonalARank4TensorCij(*args)
    def valueDesc(self):
        return "A <link linkend='Object:TrigonalARank4TensorCij'><classname>TrigonalARank4TensorCij</classname></link> object."

xmlmenudump.XMLObjectDoc(
    'TrigonalARank4TensorCij',
    xmlmenudump.loadFile('DISCUSSIONS/engine/IO/object-trigonalArank4.xml'))


class TrigonalBRank4TensorCij(Comparable):
    def __init__(self,c11,c12,c13,c33,c44,c14):
        self.c11 = c11
        self.c12 = c12
        self.c13 = c13
        self.c33 = c33
        self.c44 = c44
        self.c14 = c14
    def __repr__(self):
        return "TrigonalBRank4TensorCij(c11=%g, c12=%g, c13=%g, c33=%g, c44=%g, c14=%g)" % \
               (self.c11, self.c12, self.c13, self.c33, self.c44,
                self.c14)

utils.OOFdefine('TrigonalBRank4TensorCij', TrigonalBRank4TensorCij)

# Parameter for this class.
class TrigonalBCijklParameter(parameter.ParameterBase):
    def __init__(self, name, value=None, default=None, tip=None):
        self.checker = parameter.TypeChecker(TrigonalBRank4TensorCij)
        parameter.ParameterBase.__init__(self, name,
                                         value=value, default=default,
                                         tip=tip)
    packfmt = ">dddddd"
    packsize = struct.calcsize(packfmt)
    def binaryRepr(self, datafile, value):
        return struct.pack(TrigonalBCijklParameter.packfmt,
                           value.c11, value.c12, value.c13,
                           value.c33, value.c44, value.c14)
    def binaryRead(self, parser):
        b = parser.getBytes(TrigonalBCijklParameter.packsize)
        args = struct.unpack(TrigonalBCijklParameter.packfmt, b)
        return TrigonalBRank4TensorCij(*args) 
    def valueDesc(self):
        return "A <link linkend='Object:TrigonalBRank4TensorCij'><classname>TrigonalBRank4TensorCij</classname></link> object."

xmlmenudump.XMLObjectDoc(
    'TrigonalBRank4TensorCij',
    xmlmenudump.loadFile('DISCUSSIONS/engine/IO/object-trigonalBrank4.xml'))



class OrthorhombicRank4TensorCij(Comparable):
    def __init__(self,c11,c12,c13,c22,c23,c33,c44,c55,c66):
        self.c11 = c11
        self.c12 = c12
        self.c13 = c13
        self.c22 = c22
        self.c23 = c23
        self.c33 = c33
        self.c44 = c44
        self.c55 = c55
        self.c66 = c66
    def __repr__(self):
        return "OrthorhombicRank4TensorCij(c11=%g, c12=%g, c13=%g, c22=%g, c23=%g, c33=%g, c44=%g, c55=%g, c66=%g)" % \
               (self.c11, self.c12, self.c13, self.c22, self.c23,
                self.c33, self.c44, self.c55, self.c66)

utils.OOFdefine('OrthorhombicRank4TensorCij', OrthorhombicRank4TensorCij)

# Parameter for this class.
class OrthorhombicCijklParameter(parameter.ParameterBase):
    def __init__(self, name, value=None, default=None, tip=None):
        self.checker = parameter.TypeChecker(OrthorhombicRank4TensorCij)
        parameter.ParameterBase.__init__(self, name,
                                         value=value, default=default,
                                         tip=tip)
    packfmt = ">9d"
    packsize = struct.calcsize(packfmt)
    def binaryRepr(self, datafile, value):
        return struct.pack(OrthorhombicCijklParameter.packfmt,
                           value.c11, value.c12, value.c13,
                           value.c22, value.c23, value.c33, 
                           value.c44, value.c55, value.c66)
    def binaryRead(self, parser):
        b = parser.getBytes(OrthorhombicCijklParameter.packsize)
        args = struct.unpack(OrthorhombicCijklParameter.packfmt, b)
        return OrthorhombicRank4TensorCij(*args)
    def valueDesc(self):
        return "A <link linkend='Object:OrthorhombicRank4TensorCij'><classname>OrthorhombicRank4TensorCij</classname></link> object."

xmlmenudump.XMLObjectDoc(
    'OrthorhombicRank4TensorCij',
    xmlmenudump.loadFile('DISCUSSIONS/engine/IO/object-orthorhombicrank4.xml'))
    

class MonoclinicRank4TensorCij(Comparable):
    def __init__(self,c11,c12,c13,c15,c22,c23,c25,c33,c35,c44,c46,c55,c66):
        self.c11 = c11
        self.c12 = c12
        self.c13 = c13
        self.c15 = c15
        self.c22 = c22
        self.c23 = c23
        self.c25 = c25
        self.c33 = c33
        self.c35 = c35
        self.c44 = c44
        self.c46 = c46
        self.c55 = c55
        self.c66 = c66
    def __repr__(self):
        return "MonoclinicRank4TensorCij(c11=%g, c12=%g, c13=%g, c15=%g, c22=%g, c23=%g, c25=%g, c33=%g, c35=%g, c44=%g, c46=%g, c55=%g, c66=%g)" % \
               (self.c11, self.c12, self.c13, self.c15, self.c22,
                self.c23, self.c25, self.c33, self.c35, self.c44,
                self.c46, self.c55, self.c66)

utils.OOFdefine('MonoclinicRank4TensorCij', MonoclinicRank4TensorCij)

# Parameter for this class.
class MonoclinicCijklParameter(parameter.ParameterBase):
    def __init__(self, name, value=None, default=None, tip=None):
        self.checker = parameter.TypeChecker(MonoclinicRank4TensorCij)
        parameter.ParameterBase.__init__(self, name,
                                         value=value, default=default,
                                         tip=tip)
    packfmt = ">13d"
    packsize = struct.calcsize(packfmt)
    def binaryRepr(self, datafile, value):
        return struct.pack(MonoclinicCijklParameter.packfmt,
                           value.c11, value.c12, value.c13, value.c15,
                           value.c22, value.c23, value.c25, value.c33,
                           value.c35, value.c44, value.c46, value.c55,
                           value.c66)
    def binaryRead(self, parser):
        b = parser.getBytes(MonoclinicCijklParameter.packsize)
        args = struct.unpack(MonoclinicCijklParameter.packfmt, b)
        return MonoclinicRank4TensorCij(*args)
    def valueDesc(self):
        return "A <link linkend='Object:MonoclinicRank4TensorCij'><classname>MonoclinicRank4TensorCij</classname></link> object."

xmlmenudump.XMLObjectDoc(
    'MonoclinicRank4TensorCij',
    xmlmenudump.loadFile('DISCUSSIONS/engine/IO/object-monoclinicrank4.xml'))


# The general case.
class TriclinicRank4TensorCij(Comparable):
    def __init__(self,c11,c12,c13,c14,c15,c16,c22,c23,c24,c25,c26,
                 c33,c34,c35,c36,c44,c45,c46,c55,c56,c66):
        self.c11 = c11
        self.c12 = c12
        self.c13 = c13
        self.c14 = c14
        self.c15 = c15
        self.c16 = c16
        self.c22 = c22
        self.c23 = c23
        self.c24 = c24
        self.c25 = c25
        self.c26 = c26
        self.c33 = c33
        self.c34 = c34
        self.c35 = c35
        self.c36 = c36
        self.c44 = c44
        self.c45 = c45
        self.c46 = c46
        self.c55 = c55
        self.c56 = c56
        self.c66 = c66
    def __repr__(self):
        return "TriclinicRank4TensorCij(c11=%g, c12=%g, c13=%g, c14=%g, c15=%g, c16=%g, c22=%g, c23=%g, c24=%g, c25=%g, c26=%g, c33=%g, c34=%g, c35=%g, c36=%g, c44=%g, c45=%g, c46=%g, c55=%g, c56=%g, c66=%g)" % \
               (self.c11, self.c12, self.c13, self.c14, self.c15, self.c16,
                self.c22, self.c23, self.c24, self.c25, self.c26,
                self.c33, self.c34, self.c35, self.c36,
                self.c44, self.c45, self.c46, self.c55, self.c56, self.c66)

utils.OOFdefine('TriclinicRank4TensorCij', TriclinicRank4TensorCij)

# Parameter for this class.
class TriclinicCijklParameter(parameter.ParameterBase):
    def __init__(self, name, value=None, default=None, tip=None):
        self.checker = parameter.TypeChecker(TriclinicRank4TensorCij)
        parameter.ParameterBase.__init__(self, name,
                                         value=value, default=default,
                                         tip=tip)
    packfmt = ">21d"
    packsize = struct.calcsize(packfmt)
    def binaryRepr(self, datafile, value):
        return struct.pack(TriclinicCijklParameter.packfmt,
                           value.c11, value.c12, value.c13, value.c14,
                           value.c15, value.c16, value.c22, value.c23,
                           value.c24, value.c25, value.c26, value.c33,
                           value.c34, value.c35, value.c36, value.c44,
                           value.c45, value.c46, value.c55, value.c56, 
                           value.c66)
    def binaryRead(self, parser):
        b = parser.getBytes(TriclinicCijklParameter.packsize)
        args = struct.unpack(TriclinicCijklParameter.packfmt, b)
        return TriclinicRank4TensorCij(*args) 
    def valueDesc(self):
        return "A <link linkend='Object:TriclinicRank4TensorCij'><classname>TriclinicRank4TensorCij</classname></link> object."


xmlmenudump.XMLObjectDoc(
    'TriclinicRank4TensorCij',
    xmlmenudump.loadFile('DISCUSSIONS/engine/IO/object-triclinicrank4.xml'))
