# -*- python -*-
# $RCSfile: skeletonIO.py,v $
# $Revision: 1.28 $
# $Author: langer $
# $Date: 2005/02/24 20:39:40 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov. 

# Menu commands for reading a Skeleton from a file.

# TODO LATER: the name of the Skeleton appears many times in the file,
# in each menu command.  That makes it hard for the user to edit the
# file to change the name.  Ugly solution: have a global
# "currentskeleton" variable which is set by the initial Skeleton.New
# command.  Difficult solution: allow variables to be set locally *in*
# the data file, and store the skeleton name there.

from common.IO import mainmenu
from common.IO import oofmenu
from common.IO import parameter
from common.IO import whoville
from engine import skeletoncontext
from engine import skeletonboundary
import common.microstructure
import engine.skeleton
from common import debug
from SWIG.common import switchboard

OOFMenuItem = oofmenu.OOFMenuItem

OOF = mainmenu.OOF

skelmenu = OOF.LoadData.addItem(OOFMenuItem('Skeleton'))

def _newSkel(menuitem, name, microstructure):
    engine.skeleton.newEmptySkeleton(name, microstructure)
    
skelmenu.addItem(OOFMenuItem(
    'New',
    callback=_newSkel,
    params=[parameter.StringParameter('name', tip="Name for the Skeleton."),
            whoville.WhoParameter('microstructure',
                         common.microstructure.microStructures,
                                  tip=parameter.emptyTipString)],
    help="Load Skeleton. Used internally in a data file.",
    discussion="<para>Initiate loading &skel; from a data file.</para>"))

###

def _loadNodes(menuitem, skeleton, points):
    # read nodes as (x,y) tuples of floats
    skelcontext = skeletoncontext.skeletonContexts[skeleton]
    skeleton = skelcontext.getObject()
    for node in points:
        skeleton.newNode(*node)
    switchboard.notify(('who changed', 'Skeleton'), skelcontext)

skelmenu.addItem(OOFMenuItem(
    'Nodes',
    callback=_loadNodes,
    params=[whoville.WhoParameter('skeleton', skeletoncontext.skeletonContexts,
                                  tip=parameter.emptyTipString),
            parameter.ListOfTuplesOfFloatsParameter('points',
                                                    tip="List of points (nodes).")],
    help="Load Nodes. Used internally in a data file.",
    discussion="<para>Load <link linkend='Section:Concepts:Skeleton:Node'>nodes</link> from a <link linkend='MenuItem:OOF.File.Save.Skeleton'>saved</link> Skeleton.</para>"
    ))

###

def _loadElements(menuitem, skeleton, nodes):
    # read elements as tuples of node indices
    skelcontext = skeletoncontext.skeletonContexts[skeleton]
    skeleton = skelcontext.getObject()
    for nodelist in nodes:
        skeleton.loadElement(*nodelist)
    skelcontext.getTimeStamp(None).increment()
    switchboard.notify(('who changed', 'Skeleton'), skelcontext)

skelmenu.addItem(OOFMenuItem(
    'Elements',
    callback=_loadElements,
    params=[whoville.WhoParameter('skeleton', skeletoncontext.skeletonContexts,
                                  tip=parameter.emptyTipString),
            parameter.ListOfTuplesOfIntsParameter('nodes',
                                                  tip="List of element connectivities (List of node indices).")],
    help="Load Elements. Used internally in a data file.",
    discussion="<para>Load <link linkend='Section:Concepts:Skeleton:Element'>elements</link> from a <link linkend='MenuItem:OOF.File.Save.Skeleton'>saved</link> Skeleton.</para>"
    ))

###

def _loadPinnedNodes(menuitem, skeleton, nodes):
    # "nodes" is a list of integer indices into skeleton.nodes
    skelcontext = skeletoncontext.skeletonContexts[skeleton]
    skeleton = skelcontext.getObject()
    pinned = skelcontext.pinnednodes
    pinned.pin([skeleton.nodes[i] for i in nodes])
    pinned.signal()
skelmenu.addItem(OOFMenuItem(
    'PinnedNodes',
    callback = _loadPinnedNodes,
    params = [
    whoville.WhoParameter('skeleton', skeletoncontext.skeletonContexts,
                          tip=parameter.emptyTipString),
    parameter.ListOfIntsParameter('nodes', tip="List of indices of pinned nodes.")],
    help="Load pinned Nodes. Used internally in a data file.",
    discussion="<para>Load <link linkend='MenuItem:OOF.Graphics_n.Toolbox.Pin_Nodes.Pin'>pinned</link> nodes from a &skel; data file.</para>"
    ))

###
    
def _loadPointBoundary(menuitem, skeleton, name, nodes, exterior):
    # "nodes" is a list of integer indices into skeleton.nodes
    skelcontext = skeletoncontext.skeletonContexts[skeleton]
    skeleton = skelcontext.getObject()
    skelcontext.createPointBoundary(name, [skeleton.nodes[i] for i in nodes],
                                    exterior, autoselect=0)
    # autoselect=0 prevents the boundary from being selected in the
    # skelcontext. Selection shouldn't happen except by explicit user
    # action.

skelmenu.addItem(OOFMenuItem(
    'PointBoundary',
    callback=_loadPointBoundary,
    params=[
    whoville.WhoParameter('skeleton', skeletoncontext.skeletonContexts,
                          tip=parameter.emptyTipString),
    parameter.StringParameter('name', tip="Name of Point Boundary."),
    parameter.ListOfIntsParameter('nodes', tip="List of node indices."),
    parameter.IntParameter('exterior', 0, tip="1 (true) for the exterior boundary and 0 (false) for otherwise.")],
    help="Load Point Boundary. Used internally in a data file.",
    discussion="<para>Load a <link linkend='Section:Concepts:Skeleton:Boundary:Point'><classname>PointBoundary</classname></link> from a Skeleton data file.</para>"
    ))

###

def _loadEdgeBoundary(menuitem, skeleton, name, edges, exterior):
    # "edges" is a list of pairs of integer indices into skeleton.nodes
    skelcontext = skeletoncontext.skeletonContexts[skeleton]
    skeleton = skelcontext.getObject()
    seglist = [skeleton.getSegment(
        *[skeleton.nodes[i] for i in e] )
               for e in edges]
    startnode = skeleton.nodes[edges[0][0]]
    skelcontext.createEdgeBoundary(name, seglist, startnode,
                                   exterior, autoselect=0)
    # autoselect=0 prevents the boundary from being selected in the
    # skelcontext. Selection shouldn't happen except by explicit user
    # action.

    
skelmenu.addItem(OOFMenuItem(
    'EdgeBoundary',
    callback=_loadEdgeBoundary,
    params=[
    whoville.WhoParameter('skeleton', skeletoncontext.skeletonContexts,
                          tip=parameter.emptyTipString),
    parameter.StringParameter('name', tip="Name of Edge Boundary."),
    parameter.ListOfTuplesOfIntsParameter('edges',
                                          tip="List of Edges -- tuple of two nodes."),
    parameter.IntParameter('exterior', 0, tip="1 (true) for the exterior boundary and 0 (false) for otherwise.")],
    help="Load Edge Boundary. Used internally in a data file.",
    discussion="<para>Load a <link linkend='Section:Concepts:Skeleton:Boundary:Edge'><classname>EdgeBoundary</classname></link> from a Skeleton data file.</para>"
    ))

######

def _loadNodeGroup(menuitem, skeleton, name, nodes):
    # nodes is a list of integer indices into skeleton.nodes.
    skelcontext = skeletoncontext.skeletonContexts[skeleton]
    skeleton = skelcontext.getObject()
    groupset = skelcontext.nodegroups
    groupset.addGroup(name)
    groupset.addToGroup(name, [skeleton.nodes[i] for i in nodes])

skelmenu.addItem(OOFMenuItem(
    'NodeGroup',
    callback=_loadNodeGroup,
    params=[
    whoville.WhoParameter('skeleton', skeletoncontext.skeletonContexts,
                          tip=parameter.emptyTipString),
    parameter.StringParameter('name', tip="Name for the node group."),
    parameter.ListOfIntsParameter('nodes', tip="List of node indices.")],
    help="Load Node Group. Used internally in a data file.",
    discussion="<para>Load a node <link linkend='Section:Concepts:Skeleton:Groups'>group</link>.</para>"
    ))

def _loadElementGroup(menuitem, skeleton, name, elements):
    # elements is a list of integer indices into skeleton.elements.
    skelcontext = skeletoncontext.skeletonContexts[skeleton]
    skeleton = skelcontext.getObject()
    groupset = skelcontext.elementgroups
    groupset.addGroup(name)
    groupset.addToGroup(name, [skeleton.elements[i] for i in elements])

skelmenu.addItem(OOFMenuItem(
    'ElementGroup',
    callback=_loadElementGroup,
    params=[
    whoville.WhoParameter('skeleton', skeletoncontext.skeletonContexts,
                          tip=parameter.emptyTipString),
    parameter.StringParameter('name', tip="Name for the element group."),
    parameter.ListOfIntsParameter('elements', tip="List of element indices")],
    help="Load Element Group. Used internally in a data file.",
    discussion="<para>Load an element <link linkend='Section:Concepts:Skeleton:Groups'>group</link>.</para>"
    ))

def _loadSegmentGroup(menuitem, skeleton, name, segments):
    # segments is a list of tuples of integer indices into skeleton.nodes.
    skelcontext = skeletoncontext.skeletonContexts[skeleton]
    skeleton = skelcontext.getObject()
    groupset = skelcontext.segmentgroups
    groupset.addGroup(name)
    groupset.addToGroup(name, [skeleton.getSegment(skeleton.nodes[i],
                                                   skeleton.nodes[j])
                               for (i,j) in segments])

skelmenu.addItem(OOFMenuItem(
    'SegmentGroup',
    callback=_loadSegmentGroup,
    params=[
    whoville.WhoParameter('skeleton', skeletoncontext.skeletonContexts,
                          tip=parameter.emptyTipString),
    parameter.StringParameter('name', tip="Name for the segment group."),
    parameter.ListOfTuplesOfIntsParameter('segments', tip="List of segments -- tuple of two node indices.")],
    help="Load Segment Group. Used internally in a data file.",
    discussion="<para>Load a segment <link linkend='Section:Concepts:Skeleton:Groups'>group</link>.</para>"
    ))


#############

def writeSkeleton(datafile, skelcontext):
    skelcontext.begin_reading()
    try:
        skeleton = skelcontext.getObject()
        skelpath = skelcontext.path()

        # Create skeleton.
        datafile.startCmd(skelmenu.New)
        datafile.argument('name', skelcontext.name())
        datafile.argument('microstructure', skeleton.MS.name())
        datafile.endCmd()

        # Define nodes.
        datafile.startCmd(skelmenu.Nodes)
        datafile.argument('skeleton', skelpath)
        datafile.argument('points', [(nd.position().x, nd.position().y)
                                     for nd in skeleton.nodes])
        datafile.endCmd()

        # Nodes are written and read in the order in which they're stored
        # in the Skeleton's nodes list.  This is not the same as the
        # nodes' internal index, so we need to create a dictionary of
        # indices in the nodes list so that the elements' node tuples are
        # correct.
        nodedict = {}
        c = 0
        for node in skeleton.nodes:
            nodedict[node] = c
            c += 1
        # same for elements
        elementdict = {}
        c = 0
        for element in skeleton.elements:
            elementdict[element] = c
            c += 1

        # Define elements, using the node indices in nodedict.
        datafile.startCmd(skelmenu.Elements)
        datafile.argument('skeleton', skelpath)
        datafile.argument('nodes', [tuple([nodedict[node] for node in el.nodes])
                                    for el in skeleton.elements])
        datafile.endCmd()

        # Node groups
        for group in skelcontext.nodegroups.groups:
            datafile.startCmd(skelmenu.NodeGroup)
            datafile.argument('skeleton', skelpath)
            datafile.argument('name', group)
            datafile.argument('nodes', [nodedict[node] for node in skelcontext.nodegroups.get_group(group)])
            datafile.endCmd()

        # Element groups
        for group in skelcontext.elementgroups.groups:
            datafile.startCmd(skelmenu.ElementGroup)
            datafile.argument('skeleton', skelpath)
            datafile.argument('name', group)
            datafile.argument(
                'elements',
                [elementdict[el]
                 for el in skelcontext.elementgroups.get_group(group)])
            datafile.endCmd()

        for group in skelcontext.segmentgroups.groups:
            datafile.startCmd(skelmenu.SegmentGroup)
            datafile.argument('skeleton', skelpath)
            datafile.argument('name', group)
            nodepairs = [seg.nodes()
                         for seg in skelcontext.segmentgroups.get_group(group)]
            datafile.argument('segments',
                              [(nodedict[n1], nodedict[n2]) for (n1, n2) in nodepairs])
            datafile.endCmd()

        # Pinned nodes
        datafile.startCmd(skelmenu.PinnedNodes)
        datafile.argument('skeleton', skelpath)
        datafile.argument('nodes', [nodedict[node] for node in skelcontext.pinnednodes.retrieve()])
        datafile.endCmd()

        # Point boundaries
        for pbname, pbdy in skeleton.pointboundaries.items():
            datafile.startCmd(skelmenu.PointBoundary)
            datafile.argument('skeleton', skelpath)
            datafile.argument('name', pbname)
            datafile.argument('nodes', [nodedict[node] for node in pbdy.nodes])
            exterior = 0
            if isinstance(pbdy, skeletonboundary.ExteriorSkeletonPointBoundary):
                exterior = 1
            datafile.argument('exterior', exterior)
            datafile.endCmd()

        # Edge boundaries
        for ebname, ebdy in skeleton.edgeboundaries.items():
            datafile.startCmd(skelmenu.EdgeBoundary)
            datafile.argument('skeleton', skelpath)
            datafile.argument('name', ebname)
            edgeset = rearrangeEdges([
                tuple([nodedict[x] for x in edge.get_nodes()])
                for edge in ebdy.edges
                ])
            datafile.argument('edges', edgeset)
            exterior = 0
            if isinstance(ebdy, skeletonboundary.ExteriorSkeletonEdgeBoundary):
                exterior = 1
            datafile.argument('exterior', exterior)
            datafile.endCmd()
    finally:
        skelcontext.end_reading()
        
###########################################################

# A function to put edges in a right order.
def rearrangeEdges(edges):
    # 'edges' is a list of tuples of integers.  Each tuple is an edge,
    # and each integer is a node index (from nodedict in
    # writeSkeleton, above).
    if len(edges) == 1:
        return edges

    # Construct dictionaries of edges keyed by their starting and end points.
    tails = {}           
    heads = {}         
    for edge in edges:
        heads[edge[0]] = edge
        tails[edge[1]] = edge

    newedges = [edges[0]]

    # Build the list of contiguous edges going backwards from the
    # start point of edges[0].
    lastpt = edges[0][0]
    try:
        while 1:
            next = tails[lastpt]
            newedges.append(next)
            lastpt = next[0]
    except KeyError:
        pass

    newedges.reverse()

    # Extend the list of contiguous edges going forwards from the end
    # point of edges[0].
    lastpt = edges[0][1]
    try:
        while 1:
            next = heads[lastpt]
            newedges.append(next)
            lastpt = next[1]
    except KeyError:
        pass

    return newedges
