# -*- python -*-
# $RCSfile: skeletoninfo.py,v $
# $Revision: 1.22 $
# $Author: langer $
# $Date: 2005/01/13 22:40:13 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov. 

from engine import skeletonelement
from engine import skeletonsegment
from engine import skeletonnode
from engine import deputy
from SWIG.common import switchboard
from common import debug
from common import primitives
from common import toolbox
from common import ringbuffer
from common.IO import oofmenu
from common.IO import parameter
from common.IO import reporter
from SWIG.common import timestamp

SkeletonElement = skeletonelement.SkeletonElement
SkeletonSegment = skeletonsegment.SkeletonSegment
SkeletonNode = skeletonnode.SkeletonNode

class SkeletonQueryContainer:
    def __init__(self, context):
        self.timestamp = timestamp.TimeStamp()
        self.context = context
        self.skeleton = context.getObject()
        self.object = None
        self.position = None
        self.targetname = None

    def set(self, object=None, targetname=None, position=None):
        self.object = object
        self.targetname = targetname
        self.position = position
        self.timestamp.increment()

    def getTimeStamp(self):
        return self.timestamp

    def clone(self):
        krusty = SkeletonQueryContainer(self.context)
        krusty.skeleton = self.skeleton
        krusty.object = self.object
        krusty.targetname = self.targetname
        krusty.position = self.position
        krusty.timestamp.increment()
        return krusty

    def clearable(self):
        return not not self.object

class SkeletonPeekContainer(SkeletonQueryContainer):
    def __init__(self, context):
        self.timestamp = timestamp.TimeStamp()
        self.context = context
        self.skeleton = context.getObject()
        self.objects = {"Element":None, "Segment":None, "Node":None}

    def assignObject(self, object, objtype):
        self.objects[objtype] = object
        self.timestamp.increment()

    def removeObject(self, objtype):
        self.objects[objtype] = None
        self.timestamp.increment()

    def reset(self):
        self.objects = {"Element":None, "Segment":None, "Node":None}
        self.timestamp.increment()

    def getTimeStamp(self):
        return self.timestamp

##################################################################
            
class SkeletonInfoToolbox(toolbox.Toolbox):
    def __init__(self, gfxwindow):
        toolbox.Toolbox.__init__(self, 'Skeleton_Info', gfxwindow)
        self.whoset = ('Skeleton',)
        self.querier = None
        self.peeker = None        
        self.records = ringbuffer.RingBuffer(49)

        self.sbcallbacks = [
            # Looks for a skeleton on the gfx window.
            switchboard.requestCallback((self.gfxwindow(), "layers changed"),
                                        self.newLayers),
            # Looks for a skeleton modification.
            switchboard.requestCallback(('who changed', 'Skeleton'),
                                        self.skelChanged)
            ]

    def close(self):
        map(switchboard.removeCallback, self.sbcallbacks)

    def makeMenu(self, menu):
        self.menu = menu
        menu.addItem(oofmenu.OOFMenuItem(
            'QueryElement',
            callback=self.queryElem,
            params=[parameter.FloatParameter('x', 0.0,
                                             tip="The x coordinate of the point."),
                    parameter.FloatParameter('y', 0.0,
                                             tip="The y coordinate of the point.")],
            help="Query the closest element to the point(x,y)."
            ))
        menu.addItem(oofmenu.OOFMenuItem(
            'QuerySegment',
            callback=self.querySgmt,
            params=[parameter.FloatParameter('x', 0.0,
                                             tip="The x coordinate of the point."),
                    parameter.FloatParameter('y', 0.0,
                                             tip="The y coordinate of the point.")],
            help="Query the closest segment to the point(x,y)."
            ))
        menu.addItem(oofmenu.OOFMenuItem(
            'QueryNode',
            callback=self.queryNode,
            params=[parameter.FloatParameter('x', 0.0,
                                             tip="The x coordinate of the point."),
                    parameter.FloatParameter('y', 0.0,
                                             tip="The y coordinate of the point.")],
            help="Query the closest node to the point(x,y)."
            ))

    def activate(self):
        # Called when this toolbox has been brought to the front.
        skelcontext = self.gfxwindow().topwho(*self.whoset)
        if skelcontext is not None:
            if not self.querier:
                self.querier = SkeletonQueryContainer(skelcontext)
            if not self.peeker:
                self.peeker = SkeletonPeekContainer(skelcontext)
        else:
            self.querier = None
            self.peeker = None

    def newLayers(self):
        skelcontext = self.gfxwindow().topwho(*self.whoset)
        if skelcontext is not None:
            if self.querier and \
               skelcontext is not self.querier.context:
                self.resetRecords()
                self.querier = SkeletonQueryContainer(skelcontext)
                self.peeker = SkeletonPeekContainer(skelcontext)
                self.redraw()
            else:
                self.querier = SkeletonQueryContainer(skelcontext)
                self.peeker = SkeletonPeekContainer(skelcontext)

    def skelChanged(self, context):
        skelcontext = self.gfxwindow().topwho(*self.whoset)
        if not skelcontext: return  # No skeleton on the gfx window.
        if context is skelcontext and \
           isinstance(context.getObject(), deputy.DeputySkeleton):
            if self.querier:
                if context.getObject().sheriffSkeleton() is \
                   self.querier.skeleton:
                    # It should be node-only modification.
                    # Thus, it is not necessary to dump all the records but
                    # needs to reflect new node positions, if applicable.
                    self.querysignal()
                else:
                    self.resetRecords()
                    self.querier = SkeletonQueryContainer(skelcontext)
                    self.peeker = SkeletonPeekContainer(skelcontext)
                    self.querysignal()                    
            else:
                self.querier = SkeletonQueryContainer(skelcontext)
                self.peeker = SkeletonPeekContainer(skelcontext)
        else:
            # The skeleton has been modified.
            # Thus, the query records should be reset, since "prev",
            # "next" would be meaningless after the modifications such as
            # "Refine" or "Rationalize".
            self.resetRecords()
            self.querier = SkeletonQueryContainer(skelcontext)
            self.peeker = SkeletonPeekContainer(skelcontext)
            self.querysignal()
        
    def resetRecords(self):
        self.records.clear()

    def queryElem(self, menuitem, x, y):
        context = self.gfxwindow().topwho(*self.whoset)
        if not context:
            return
        point = primitives.Point(x,y)
        skeleton = context.getObject()
        elem = skeleton.enclosingElement(point)
        if not elem:
            reporter.report("Try to click ON an Element, dude.")
        else:
            self.finishQuery(elem, "Element", point)

    def querySgmt(self, menuitem, x, y):
        context = self.gfxwindow().topwho(*self.whoset)
        if not context:
            return
        point = primitives.Point(x,y)
        skeleton = context.getObject()
        sgmt = skeleton.nearestSgmt(point)
        if not sgmt:
            reporter.report("Avoid clicking OUTSIDE of a skeleton.")
        else:
            self.finishQuery(sgmt, "Segment", point)

    def queryNode(self, menuitem, x, y):
        context = self.gfxwindow().topwho(*self.whoset)
        if not context:
            return
        point = primitives.Point(x,y)
        skeleton = context.getObject()
        node = skeleton.nearestNode(point)
        if not node:
            reporter.report("Avoid clicking OUTSIDE of a skeleton.")
        else:
            self.finishQuery(node, "Node", point)

    def finishQuery(self, object, targetname, position):
        self.querier.set(object=object, targetname=targetname,
                         position=position)
        self.records.push(self.querier.clone())
        self.peeker.reset()
        self.querysignal()
        
    def querysignal(self):
        switchboard.notify((self.gfxwindow(),"query skeleton"))
        self.redraw()

    def clearQuerier(self):
        if self.querier:
            self.querier.set()
            self.peeker.reset()
            self.redraw()

    def redraw(self):
        switchboard.notify("redraw")
        
    def prev_able(self):
        return not self.records.atBottom()

    def next_able(self):
        return not self.records.atTop()

    tip="Get information about Skeleton components."
    discussion="""<para>
    Get information about &skel; components, based on mouse input.
    </para>"""
    
toolbox.registerToolboxClass(SkeletonInfoToolbox, ordering=2.0)
