# -*- python -*-
# $RCSfile: analysisdomain.py,v $
# $Revision: 1.20 $
# $Author: reida $
# $Date: 2004/12/17 18:46:49 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov. 

# Domain objects.  These are part of the post-solution mesh analysis
# process.  A "domain" is a user-specified portion of a mesh or that
# mesh's microstructure, specified in terms of mesh or microstructure
# objects, like pixels, elements, cross-sections, etc.

# One cannot actually evalute an output on a domain -- it's necessray
# to sample the domain first, and the sample objects are the
# numerically savvy components.  There can be more than one sampling
# of a domain.

from common import microstructure
from common import primitives
from common import registeredclass
from common.IO import pixelgroupparam
from common.IO import selectionplaceholder
from common.IO import whoville
from engine import skeletoncontext
from engine.IO import meshcsparams
from engine.IO import skeletongroupparams
import engine.mesh


class Domain(registeredclass.RegisteredClass):
    registry = []
    def set_mesh(self, mesh):
        self.mesh = mesh
        self.mesh_obj = engine.mesh.meshes[mesh]
        self.femesh = self.mesh_obj.getObject()
        self.skeleton = self.mesh_obj.skeleton
        self.skelcontext = skeletoncontext.skeletonContexts[
            self.skeleton.name()]
        self.ms = self.mesh_obj.getMicrostructure()
        self.mscontext = microstructure.getMSContextFromMS(self.ms)
    def read_lock(self):
        self.mesh_obj.begin_reading()
        self.skelcontext.begin_reading()
        self.mscontext.begin_reading()
    def read_release(self):
        self.mscontext.end_reading()
        self.skelcontext.end_reading()
        self.mesh_obj.end_reading()

Domain.tip = "A specified region of Mesh or its Microstructure for sampling (see <link linkend='RegisteredClass:SampleSet'><classname>SampleSet</classname></link>)."
Domain.discussion = """
<para>
Domain objects are part of the post-solution analysis framework,
in which useful data is extracted from the mesh.
You select an operation (like Direct Output, if you want to
see the data, or Average), a Domain, and a Sampling.
</para><para>
The Domain is the subset of the mesh to which the sampling is
applied.  All the samples thus generated will lie within the
domain, or in some cases within the domain's bounding box.
</para><para>
All domains are ways of specifying an area on the mesh, and
differ in how this area is specified.  You may select
an element group selection, a pixel group or selection,
a cross section, or the entire mesh.  Exactly how the
domain is interpreted by the output operation depends on the
sampling.
</para>"""


# Each subclass implements some subset of retrieval functions, which
# return information about the domain, in the form of a bounding-box
# rectangle, or a list of objects (elements, points).  The caller is
# expected to be something that will then construct sample objects, or
# wrap the returned objects in sample objects.

# Retrieval functions are things like get_bounds, get_pixels,
# get_endpoints.  Not all retrieval functions exist for all domains,
# and some retrieval functions might be expensive.

# Units, where they make sense, are physical units on the
# microstructure.

# Domains need to indicate what kinds of objects make sense for
# sampling, using the constants defined here.
CS_POINT = "cs_point_sample" # get_endpoints and get_elemental_segments.
POINT =    "point_sample"   # Promises existence of get_bounds, for the bbox.
PIXEL =    "pixel_sample"   # Promises existence of get_pixels.
ELEMENT =  "element_sample"  # Promises existence of get_elements.


# An object for operating on the entire mesh.
class EntireMesh(Domain):
    # Returns a bounding rectangle.
    def get_bounds(self):
        return primitives.Rectangle(primitives.Point(0.0,0.0),
                                    self.ms.size())
    
    # Returns a list of iPoint pixel indices.
    def get_pixels(self):
        return self.ms.coords()

    # Returns a list of all the elements.  This maybe should return the
    # iterator, to save memory?  The number of elements could be large.
    def get_elements(self):
        el_list = []
        iter = self.femesh.element_iterator()
        while not iter.end():
            el_list.append(iter.element())
            iter.next()
        return el_list
        

registeredclass.Registration(
    'Entire Mesh', Domain, EntireMesh, 0, params=[],
    sample_types = [POINT, PIXEL, ELEMENT] )

# Pixel selection or group -- these actually live on
# the host microstructure, so you have to go there to get 'em.
class PixelGroup(Domain):
    def __init__(self, pixels):
        self.pixels = pixels
    def _pixels_from_group(self):
        if self.pixels==selectionplaceholder.selectionPlaceHolder:
            group = self.ms.pixelselection.getSelectionAsGroup()
        else:
            group = self.ms.findGroup(self.pixels)
        if group:
            return group.members()
    def get_pixels(self):
        return self._pixels_from_group()
    def get_bounds(self):
        ms_size = self.ms.size()
        pixels = self._pixels_from_group()

        maxx=0.0
        maxy=0.0
        minx=ms_size[0]
        miny=ms_size[1]
        pixel_size = self.ms.sizeOfPixels()

        for p in pixels:
            low_x = p[0]*pixel_size[0]
            hi_x = low_x + pixel_size[0]
            low_y = p[1]*pixel_size[1]
            hi_y = low_y + pixel_size[1]
            minx = min(minx, low_x)
            miny = min(miny, low_y)
            maxx = max(maxx, hi_x)
            maxy = max(maxy, hi_y)

        return primitives.Rectangle(primitives.Point(minx, miny),
                                    primitives.Point(maxx, maxy))
    
registeredclass.Registration(
    'Pixel Group', Domain, PixelGroup, 10,
    params=[pixelgroupparam.PixelAggregateParameter('pixels')],
    sample_types = [POINT, PIXEL] )

# Cross-section -- currently, all cross-sections are straight, so the
# endpoints provide a complete description.  Also, the elemental
# segments are straight.  This is named "CrossSectionDomain" to
# distinguish it from the CrossSection class which represents actual
# cross-sections in the mesh.
class CrossSectionDomain(Domain):
    def __init__(self, cross_section):
        self.cross_section=cross_section
    def get_endpoints(self):
        cs_obj = self.mesh_obj.getCrossSection(self.cross_section)
        return (cs_obj.start, cs_obj.end)

    # Builds (and returns) a list of tuples, each consisting of a
    # primitives.Segment object and a mesh element.  These occur in
    # order from start to end, which map exactly once on to the cross
    # section, and whose boundaries correspond to intersections of the
    # cross-section with element boundaries.
    def get_elemental_segments(self):
        cs_obj = self.mesh_obj.getCrossSection(self.cross_section)
        local_segment = primitives.Segment(cs_obj.start, cs_obj.end)
        skeleton = self.skeleton

        start_el = skeleton.enclosingElement(cs_obj.start)

        result = []
        # "last" means "most recent", and is distinct from "final".
        last_pt = cs_obj.start
        last_el = start_el
        (isec, new_el) = \
               skeleton.get_intersection_and_next_element(
            local_segment, last_el)

        # If the *first* call yields no intersections, then the
        # whole cross-section lies within one element.  Return
        # a single segment made up of the whole cross-section.
        if not isec:
            result.append( (primitives.Segment(last_pt, cs_obj.end),
                            self.femesh.getElement(last_el.meshindex))  )
            return result

        # Otherwise, keep iterating until you get to the end.
        while new_el:
            result.append( (primitives.Segment(last_pt, isec),
                            self.femesh.getElement(last_el.meshindex))  )
            last_pt = isec
            last_el = new_el
            (isec, new_el) = \
                   skeleton.get_intersection_and_next_element(
                local_segment, last_el)

        # Bolt on the final one.
        result.append( (primitives.Segment(last_pt, cs_obj.end),
                        self.femesh.getElement(last_el.meshindex))  )

        return result


registeredclass.Registration(
    'Cross Section', Domain, CrossSectionDomain, 20,
    params=[meshcsparams.MeshCrossSectionParameter('cross_section')],
    sample_types = [CS_POINT])


class ElementGroup(Domain):
    def __init__(self, elements):
        self.elements = elements
    def _elements_from_aggregate(self):
        skeleton = self.skeleton
        skel_els = []
        if self.elements == selectionplaceholder.selectionPlaceHolder:
            skel_els = self.skelcontext.elementselection.retrieveFromSkeleton(
                skeleton)
        else:
            skel_els = self.skelcontext.elementgroups.get_groupFromSkeleton(
                self.elements, skeleton)
        return [self.femesh.getElement(s.meshindex) for s in skel_els]

    def get_bounds(self):
        first = 1
        for e in self._elements_from_aggregate():
            for n in e.cornernode_iterator():
                pos = n.position()
                if first:
                    xmin = pos[0]
                    xmax = pos[0]
                    ymin = pos[1]
                    ymax = pos[1]
                    first = None
                else:
                    xmin = min(pos[0], xmin)
                    xmax = max(pos[0], xmax)
                    ymin = min(pos[1], ymin)
                    ymax = max(pos[1], ymax)
        return primitives.Rectangle(primitives.Point(xmin, ymin),
                                    primitives.Point(xmax, ymax) )
                
    def get_elements(self):
        return self._elements_from_aggregate()

registeredclass.Registration(
    'Element Group', Domain, ElementGroup, 30,
    params=[skeletongroupparams.ElementAggregateParameter('elements')],
    sample_types=[POINT, ELEMENT])

