# -*- python -*-
# $RCSfile: analysissample.py,v $
# $Revision: 1.25 $
# $Author: reida $
# $Date: 2005/01/31 21:14:08 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov. 

# Sampling objects.  A sampling object applies to a particular domain, and
# upon invocation (with appropriate parameters), returns a list of
# samples.  In principle, samples can answer quantitative questions
# about themselves, such as:
#  - What is your location or identifier?
#  - What is the value of output O(location)^n on you?
#  - (advanced) What fraction of you satisfies x < O(location) < y?
#
# Element samples understand "value" to mean "integral".  The area of
# an element can be found by evaluating an output to the zeroth power.
# 
#   With that caveat, all of the analysis operations should be able to
# compute themselves by asking questions of this type of their
# samples.  For efficiency, samples can cache expensive data, such as
# their area or important area-fractions.

# Some samples make sense to evaluate directly, like points on a grid
# or points on a line, and others do not, because they have no
# fixed location, like elements and pixels.
#   Directly-evaluate-able sample sets should register themselves
# with "direct=1", and others should not.

from common import primitives
from common import registeredclass
from common.IO import automatic
from common.IO import parameter
from engine import analysisdomain
from SWIG.engine import outputval
import math
import string
import types

# Sample objects should have a "columnNames" routine, which returns a
# list of strings which identify the columns of data which are
# provided, and a "columnData" routine, which provides a list of
# strings which make up this data.

# The base class has the output-wrapping "columnData" routine -- this
# takes as input a list of strings identifying the headers that are
# actually used, and picks those items out of the data result from the
# subclass's outputData routine.  This base class routine returns a
# list-enclosed list -- subclasses which need to return multiple data
# sets can return several lists.  The length of the top-level list
# returned from this function is the number of data sets.

# The "integrate" routine takes an argument "otype", which is the
# output type of the output -- this is passed in because the check is
# relatively expensive.  It's required because different otypes do
# arithmetic in different ways, and polymorphism of outputs is
# valuable.  Non-float outputs do component-wise exponentiation
# thorugh a "component_pow" routine, and also have special routines
# "component_square" and "component_sqrt" for the special powers of
# 2 and 1/2.


class Sample:
    def columnData(self, header):
        datalist = self.outputData()
        return [ [ datalist[self.columnNames.index(x)] for x in header] ]

class ElementSample(Sample):
    columnNames = ["Element"]
    def __init__(self, element):
        self.element = element
        self.index = element.get_index()
    def identifier(self):
        return self.index
    def integrate(self, domain, output, otype, order, power=1):
        femesh = domain.mesh_obj.getObject()
        gauss_pts =self.element.integration_points(order)
        pts = [x.mastercoord() for x in gauss_pts]
        wgts = [x.weight() for x in gauss_pts]
        res = 0.0
        if power==0:
            res = reduce(lambda x,y: x+y, wgts)
        else:
            vals = output.evaluate(femesh, [self.element], [pts])
            if power==1:
                res = reduce( lambda x,y: x+y[0]*y[1],
                              zip(wgts, vals), 0.0)
            else:
                newvals = []
                if otype == types.FloatType:
                    for v in vals:
                        newvals.append(v**power)
                elif otype == outputval.OutputValPtr:
                    for v in vals:
                        v.component_pow(power)
                        newvals.append(v)
                        
                # This can be de-stupidized once we're safely past
                # Python 2.1.  The issue is that if you make a list
                # comprehension or lambda which uses "power", Py2.1
                # assumes "power" is a global, and complains because
                # it has no value.
                res = reduce( lambda x,y: x+y[0]*y[1],
                              zip(wgts, newvals), 0.0)
        return res


    # The ElementSample is actually not used for direct output, but
    # can use this mechanism for debugging purposes.
    def outputData(self):
        return ["Element %d" % self.index]
        


class PointSample(Sample):
    columnNames = ["X", "Y"]
    def __init__(self, point):
        self.point = point
    def identifier(self):
        return self.point
    def outputData(self):
        return [`self.point.x`, `self.point.y`]




# Line samples know how far along the line they are, from 0 to 1.
# as well as in physical units.
class LineSample(PointSample):
    columnNames = ["Distance", "Fraction", "X", "Y"]
    def __init__(self, point, fraction, distance):
        PointSample.__init__(self, point)
        self.fraction = fraction
        self.distance = distance
    def identifier(self):
        return (self.point, self.fraction, self.distance)
    def outputData(self):
        return [`self.distance`, `self.fraction`,
                `self.point.x`, `self.point.y`] 


# A portion of a line contained within one element.  "Segment" is a
# primitives.Segment object, and "element" is an FEMesh element.
class ElementLineSample(Sample):
    columnNames = ["Segment", "Distance", "Fraction", "X", "Y"]
    def __init__(self, segment, element, fractions, distances, n_points):
        self.segment = segment
        self.element = element
        # "Fractions" are the fractional distance along the segment
        # at the start and end of this element; distances are
        # similar; n_points is how many points to evaluate at.
        self.fractions = fractions
        self.distances = distances
        self.n_points = n_points
    def identifier(self):
        return (self.segment, self.element)
    # Because there are multiple points, this class has a local
    # evaluate routine.  Returns a tuple of length n_points>2,
    # of evenly-spaced points.  
    def evaluate(self, femesh, output):
        start = self.segment.start()
        end = self.segment.end()
        dx = (end-start)/float(self.n_points-1)
        master_point_set = []
        for i in range(self.n_points):
            lab_point = start+i*dx
            master_point_set.append(
                self.element.to_master(lab_point[0], lab_point[1]))
        return output.evaluate( femesh, [self.element],
                                [master_point_set] )
    # This class has multiple data sets, and so overrides the
    # base-class columnData routine.
    def columnData(self, header):
        out_list = []
        start = self.segment.start()
        end = self.segment.end()
        dx = (end-start)/float(self.n_points-1)
        df = (self.fractions[1]-self.fractions[0])/float(self.n_points-1)
        dd = (self.distances[1]-self.distances[0])/float(self.n_points-1)
        for i in range(self.n_points):
            lab_pt = start+i*dx
            outputData = [`self.segment`, `self.distances[0]+i*dd`,
                          `self.fractions[0]+i*df`, `lab_pt[0]`, `lab_pt[1]`]
            out_list.append( [ outputData[self.columnNames.index(x)]
                              for x in header] )
        return out_list

# Pixels.  PixelSamples can be evaluated at their centers, or
# integrated over.  
class PixelSample(Sample):
    columnNames = ["Pixel", "X", "Y"]
    def __init__(self, pixel, size):
        # "pixel" is an iPoint pixel index.
        self.pixel = pixel
        pt_x = float(self.pixel[0])*size[0] + 0.5*size[0]
        pt_y = float(self.pixel[1])*size[1] + 0.5*size[1]
        self.point = primitives.Point(pt_x, pt_y)
    def identifier(self):
        return self.pixel
    def outputData(self):
        return [ "(%d %d)" % (self.pixel[0], self.pixel[1]),
                 `self.point[0]`, `self.point[1]` ]
    # For integration and direct output, pixels are treated as points
    # at their centers.  It may be desirable, at some point, to take
    # the integration operation more seriously.
    def integrate(self, domain, output, otype, power=1):
        if power==0:
            return 1
        v = self.evaluate(domain, output)
        if power==1:
            return v
        if otype == types.FloatType:
            return v**power
        if otype == outputval.OutputValPtr:
            return v.compoment_pow(power)
        # float and outputvalptr are the only two types allowed.
    
    # Evaluate always evaluates at the center, and returns a float.
    def evaluate(self, domain, output):
        skel_el = domain.skeleton.enclosingElement(self.point)
        real_el = domain.femesh.getElement(skel_el.meshindex)
        vlist = output.evaluate(
            domain.femesh,
            [real_el],
            [ [ real_el.to_master(self.point[0], self.point[1]) ] ] )
        return vlist[0]




#################################################################
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
#   #   #   #   #   #   #   #   #   #   #   #   #   #   #   #   #
#       #       #       #       #       #       #       #       # 


        
# Utility function, for encapsulating the transformation from column
# names to attribute names.
def _attr_from_colname(colname):
    return "show_"+string.lower(colname)



# SampleSet objects are containers for many samples, and have
# iterator-like semantics.  SampleSet evaluate routines must
# take the "otype" argument after the output, though many of them
# discard it.

class SampleSet(registeredclass.RegisteredClass):
    registry = []
    # The kwargs passed in here will be column names, with
    # boolean attributes.
    def __init__(self, **kwargs):
        self.sample_list = []
        for (attr, val) in kwargs.items():
            self.__dict__[attr]=val
    def __len__(self):
        return len(self.sample_list)
    def __getitem__(self, idx):
        return self.sample_list[idx]
    # Make-samples returns a list of actually-requested (by the user)
    # column names.  Only called for operations that need them --
    # other operations can call "fill_sample_list(domain)" directly.
    def make_samples(self, domain):
        self.fill_sample_list(domain)
        col_names = [x for x in self.containedClass.columnNames
                     if self.__dict__[_attr_from_colname(x)] ]
        return col_names

SampleSet.tip = "Container for samples. Outputs are evaluated at samples."
SampleSet.discussion = """
<para>

</para>"""


# Set of elements.
class ElementSampleSet(SampleSet):
    containedClass = ElementSample
    def __init__(self, order, **kwargs):
        self.order = order
        SampleSet.__init__(self, **kwargs)
    def fill_sample_list(self, domain):
        for e in domain.get_elements():
            self.sample_list.append(ElementSample(e))
    def integrate(self, domain, output, otype, power=1):
        if self.order==automatic.automatic:
            order=2
        else:
            order = self.order
        vals = [x.integrate(domain, output, otype, order, power)
                for x in self.sample_list]
        return zip(self.sample_list, vals)


registeredclass.Registration(
    "Element Set", SampleSet, ElementSampleSet, 10,
    params=[parameter.AutoIntParameter("order", automatic.automatic,
    tip="Set the order of integration to use during this computation.")],
    sample_type=analysisdomain.ELEMENT, direct=0 )


# class DirectSampleSet(SampleSet):
#     registry = []


# Trick function which creates the registrations for SampleSet
# objects.  In the case of direct=1, it creates two registration
# objects.  You can't just pass in the two parent classes, as is done
# for more usual multi-registrations, because we want to use different
# parameters for the different cases here.
def DirectSampleSetRegistration(name, subclass, ordering, params=[],
                                secret=0, tip=None, discussion=None, **kwargs):
    # If direct is set, then register it with direct=0 and no
    # automatic extra parameters.
    class NonDirectSubclass(subclass):
        pass

    NonDirectSubclass.__name__ = "Stat"+subclass.__name__
    
    non_direct_params = params[:]
    registeredclass.Registration(
        name, SampleSet, NonDirectSubclass, ordering,
        params=non_direct_params, secret=secret, tip=tip,
        discussion=discussion, direct=0, **kwargs)

    # Now pile on the extra parameters.
    for c_name in subclass.containedClass.columnNames:
        params.append(parameter.BooleanParameter(
            _attr_from_colname(c_name), 1, default=1,
            tip="Include or exclude this data from the output.") )

    return registeredclass.Registration(
        name, SampleSet, subclass, ordering, params, secret,
        tip, discussion, direct=1, **kwargs)


# All point sample sets share an "evaluate" routine.  It returns a vector
# of tuples, each of the form (point-sample, value-at-point).  
class PointSampleSet(SampleSet):
    def evaluate(self, domain, output, otype):
        return self.integrate(domain, output, otype, power=1)
    def integrate(self, domain, output, otype, power=1):
        femesh = domain.femesh
        skeleton = domain.skeleton
        res = []
        count = 0
        # At this point, p's have to be PointSample objects.
        for p in self.sample_list:
            count += 1
            if power==0:
                res.append( (p, 1.0) )
            else:
                pt = p.point
                skel_el = skeleton.enclosingElement(pt)
                real_el = femesh.getElement(skel_el.meshindex)
                val = output.evaluate(
                    femesh, [real_el],
                    [ [ real_el.to_master(pt[0],pt[1]) ]  ])
                # "val" is a list, known to be of length 1.
                if power==1:
                    # Skip the exponent if it's not required.
                    res.append( (p, val[0]) )
                else:
                    if otype==types.FloatType:
                        res.append( (p, val[0]**power) )
                    elif otype==outputval.OutputValPtr:
                        val[0].component_pow(power)
                        res.append( (p, val[0]) )
        return res
    
    
# Grid on a box.
class GridSampleSet(PointSampleSet):
    containedClass = PointSample
    def __init__(self, x_points, y_points, **kwargs):
        self.x_points = x_points
        self.y_points = y_points
        SampleSet.__init__(self, **kwargs)
    def fill_sample_list(self, domain):
        box = domain.get_bounds()
        bxmin = box.xmin()
        bymin = box.ymin()
        dx = float(box.xmax()-bxmin)/float(self.x_points-1)
        dy = float(box.ymax()-bymin)/float(self.y_points-1)
        for j in range(self.y_points):
            for i in range(self.x_points):
                self.sample_list.append(
                    PointSample(primitives.Point(bxmin+i*dx,
                                                 bymin+j*dy)))
                   
DirectSampleSetRegistration(
    "Grid Points", GridSampleSet, 20,
    params=[parameter.IntParameter('x_points', 50,
                                   tip="Total number of points horizontally in the grid."),
            parameter.IntParameter('y_points', 50,
                                   tip="Total number of points vertically in the grid.")],
    sample_type=analysisdomain.POINT)



# Grid of given delta-x, delta-y.
class SpacedGridSampleSet(PointSampleSet):
    containedClass = PointSample
    def __init__(self, delta_x, delta_y, **kwargs):
        self.delta_x = delta_x
        self.delta_y = delta_y
        SampleSet.__init__(self, **kwargs)
    def fill_sample_list(self, domain):
        box = domain.get_bounds()
        bxmin = box.xmin()
        bymin = box.ymin()
        bxmax = box.xmax()
        bymax = box.ymax()
        (x,y)=(bxmin,bymin)
        while 1:
            self.sample_list.append(
                PointSample(primitives.Point(x,y)))
            x += self.delta_x
            if x > bxmax:
                x = bxmin
                y += self.delta_y
                if y > bymax:
                    break


DirectSampleSetRegistration(
    "Spaced Grid Points", SpacedGridSampleSet, 21,
    params=[parameter.FloatParameter('delta_x', 0.1,
                                     tip="Horizontal spacing between grid points."),
            parameter.FloatParameter('delta_y', 0.1,
                                     tip="Vertical spacing between grid points.")],
    sample_type=analysisdomain.POINT)



# Regular array of points on a line.
class LineSampleSet(PointSampleSet):
    containedClass = LineSample
    def __init__(self, n_points, **kwargs):
        self.n_points = n_points
        SampleSet.__init__(self, **kwargs)
    def fill_sample_list(self, domain):
        (start,end)=domain.get_endpoints()
        length=math.sqrt( (end-start)**2 )
        # This way of accumulating distances may have roundoff problems.
        # Should be done more sensibly.
        ds = (end-start)*(1.0/float(self.n_points-1))
        dsl = math.sqrt(ds**2)
        for i in range(self.n_points):
            local_point = start+i*ds
            distance = i*dsl
            fraction = float(distance)/float(length)
            self.sample_list.append(
                LineSample(local_point, fraction, distance))

DirectSampleSetRegistration(
    "Line Points", LineSampleSet, 30,
    params=[parameter.IntParameter('n_points', 50,
                                   tip="Number of evenly-spaced points at which to sample the cross section.")],
    sample_type=analysisdomain.CS_POINT)


class ElementSegmentSampleSet(SampleSet):
    containedClass = ElementLineSample
    def __init__(self, n_points, **kwargs):
        self.n_points = n_points
        SampleSet.__init__(self, **kwargs)
    def fill_sample_list(self, domain):
        (cs_start, cs_end) = domain.get_endpoints()
        len = math.sqrt( (cs_end-cs_start)**2 )
        for (seg, el) in domain.get_elemental_segments():
            dist1 = math.sqrt( (seg.start()-cs_start)**2 )
            dist2 = math.sqrt( (seg.end()-cs_start)**2 )
            frac1 = dist1/len
            frac2 = dist2/len
            self.sample_list.append(
                ElementLineSample(seg, el,
                                  (frac1,frac2), (dist1,dist2),
                                  self.n_points ) )
    def evaluate(self, domain, output, otype):
        res = []
        femesh = domain.femesh
        for s in self.sample_list:
            res.append( (s, s.evaluate(femesh, output)) )
        return res

    def integrate(self, domain, output, otype, power=1):
        result = []
        for s in self.sample_list:
            start = s.segment.start()
            end = s.segment.end()
            lgth = math.sqrt( (end-start)**2 )
            if power==0:
                result.append(lgth)
            else:
                rval = 0.0
                dx = lgth/float(self.n_points-1)
                vals = s.evaluate(domain.femesh, output)
                if power!=1:
                    if otype==types.FloatType:
                        vals = [x**power for x in vals]
                    elif otype==outputval.OutputValPtr:
                        for x in vals:
                            x.component_pow(power)
                for i in range(len(vals)-1):
                    rval += (vals[i]+vals[i+1])*dx/2.0
                result.append(rval)

        return zip(self.sample_list, result)

            
DirectSampleSetRegistration(
    "Element Segments", ElementSegmentSampleSet, 40,
    params=[parameter.IntRangeParameter(
    # HACK:  This range should actually be unbounded above.
    "n_points",[2,25],tip="Number of sample points per element")],
    sample_type=analysisdomain.CS_POINT)


class PixelSampleSet(SampleSet):
    containedClass = PixelSample
    def __init__(self, **kwargs):
        SampleSet.__init__(self, **kwargs)
    def fill_sample_list(self, domain):
        size = domain.ms.sizeOfPixels()
        self.sample_list = \
                         [PixelSample(x, size) for x in domain.get_pixels()]
    def evaluate(self, domain, output, otype):
        res = []
        for s in self.sample_list:
            res.append( (s, s.evaluate(domain, output)) )
        return res
        
    def integrate(self, domain, output, otype, power=1):
        vals = [x.integrate(domain, output, otype, power) 
                for x in self.sample_list]
        return zip(self.sample_list, vals)
    

DirectSampleSetRegistration(
    "Pixels", PixelSampleSet, 50,
    params=[],
    sample_type=analysisdomain.PIXEL)
