# -*- python -*-
# $RCSfile: conjugate_.py,v $
# $Revision: 1.10 $
# $Author: reida $
# $Date: 2005/01/12 16:51:02 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov.

from common import debug
from SWIG.common import lock
from SWIG.common import switchboard
from SWIG.engine import material

from common import microstructure
from common import utils
from engine import propertyregistration

import engine.mesh

class ListOfConjugatePairs:
    def __init__(self):
        self.pairs = []
        self.number_of_pairs = 0
        self.lock = lock.Lock()

    def __del__(self):
        self.lock = None
        
    def add(self, conjpair):
        self.lock.acquire()
        if conjpair not in self.pairs:
            self.pairs.append(conjpair)
            self.number_of_pairs += 1
        self.lock.release()

    def relevant_property(self, property, pair):
        is_relevant = 0
        self.lock.acquire()
        if property.is_property_type(pair.get_name()):
            is_relevant = 1
        self.lock.release()
        return is_relevant
    
    def sanity_check(self, relevant_pairs):
        ## For a list of properties, construct a temporary
        ## list of all relevant conjugate pairs and then
        ## cross-check them to see if they conflict
        self.lock.acquire()
        
        bill_of_health = 1        
        ## Loop over relevant pairs.
        ## If two pairs address the same
        ## equation and equation components
        ## but their corresponding (conjugate) fields
        ## or field components are not the same
        ## then there is no guarantee that the stiffness
        ## matrix symmetry will be preserved
        for i in range(len(relevant_pairs)):
            cpair1 = relevant_pairs[i]
            for j in range(i + 1, ):
                cpair2 = relevant_pairs[j]
                if cpair1 != cpair2 :
                    if cpair1.get_equation() == cpair2.get_equation() \
                           and cpair1.get_equation_component() == cpair2.get_equation_component():
                        ## if they share the same equation
                        ## and they share the same equation components
                        if cpair1.get_field() != cpair2.get_field() \
                               or cpair1.get_field_component() != cpair2.get_field_component():
                            ## but they are conjugate to different field components
                            ## relationship overdefined, and symmetry may not be guaranteed
                            bill_of_health = 0
                            
        self.lock.release()
        return bill_of_health ## return state of system
    

    def find_local_mapping(self, femesh):
        ## femesh is a real FEMesh
        self.lock.acquire()
        local_row = []
        local_col = []
        
        for conjugatepair in self.pairs:
            field = conjugatepair.get_field()
            fieldcomp = conjugatepair.get_field_component()
            equation = conjugatepair.get_equation()
            equationcomp = conjugatepair.get_equation_component()
            if femesh.is_active_field(field) and femesh.is_active_equation(equation):
                eqcomp = femesh.active_equation_index(equation) + equationcomp.integer()
                fcomp = femesh.defined_index(field) + fieldcomp.integer()
                local_row.append(eqcomp)
                local_col.append(fcomp)
                
                ## print local_row, local_col
                ## print equation.name()
        ## print "cleaned list ", self.trim_list(local_col)
        femesh.set_local_mapping(local_row, self.trim_list(local_col))
        self.lock.release()
    def trim_list(self, list):
        ## Must not be locked. For internal use only. Keep away of reach of adults.
        ## 
        ## trim_list creates a list of ints that preserves the
        ## relative order between the different components of
        ## the original list, but fills the holes with the next
        ## available integer so that the minimum item in the
        ## list is zero and the maximum is n-1, where n is the
        ## list size
        list_dict = {}
        cleaned_list = []
        list_copy  = list[:]
        list_copy.sort() ## on the order nlogn
        for item in range(len(list)):
            list_dict[list_copy[item]] = item
        for item in range(len(list)):
            cleaned_list.append(list_dict[list[item]])
        return cleaned_list
    
_ConjugatePairOld__init__ = ConjugatePair.__init__

def _ConjugatePair__init__(self, name, equation, eqncomp, field, fieldcomp):
    _ConjugatePairOld__init__(self, name, equation, eqncomp, field, fieldcomp)
    listofconjugatepairs.add(self)

ConjugatePair.__init__ = _ConjugatePair__init__

listofconjugatepairs = ListOfConjugatePairs()


def check_symmetry(meshName, second_argument, third_argument):
    ## Second and third arguments are NOT used.
    ## They are only included to stop the associated swichboard call
    ## from complaining.
    ##
    ## function assumes that fields and equations
    ## in relevant pairs are all active, and that
    ## there is ONE mesh per microstructure...

    ##
    ## return zero means symmetricity is NOT guaranteed
    ## return one means symmetricity is guaranteed (for only $99.99!! :) )
    ##
    ## print "checking symmetry"
    mesh = engine.mesh.meshes[meshName].getObject()
    meshWho = engine.mesh.meshes[meshName]
    ms = engine.mesh.meshes[meshName].getMicrostructure()
    global listofconjugatepairs
    materials = material.getMaterials(ms)
    relevant_pairs = {}
    
    
    property_symmetry = 1
    ## property_symmetry tells if symmetry is compromised given
    ## the fields that are currently defined AND/OR active.


    ## construction of temporary list and check symmetry from properties
    for matter in materials: ## for every material
        for property in matter.properties(): ## for every property
            property_symmetry *= property.is_symmetric(mesh)
            ## debug.fmsg(property.name(),  property_symmetry)
            for pair in listofconjugatepairs.pairs:
                ## check property is relevant to the defined AND active
                ## cojugate pairs
                property_registration = propertyregistration.AllProperties[property.name()]
                if listofconjugatepairs.relevant_property(property_registration, pair):
                    relevant_pairs[pair] = pair
    
    ## now let me tell you if you are insane...
    meshWho.set_symmetry_state(listofconjugatepairs.sanity_check(relevant_pairs.keys())*property_symmetry)
    ## debug.fmsg( meshWho.meshSymmetry)

switchboard.requestCallback("equation activated", check_symmetry)
switchboard.requestCallback("equation deactivated", check_symmetry)

switchboard.requestCallback("field activated", check_symmetry)
switchboard.requestCallback("field deactivated", check_symmetry)
