// -*- C++ -*-
// $RCSfile: femesh.C,v $
// $Revision: 1.85 $
// $Author: redwing $
// $Date: 2005/03/01 16:31:13 $


/* This software was produced by NIST, an agency of the U.S. government,
 * and by statute is not subject to copyright in the United States.
 * Recipients of this software assume all responsibilities associated
 * with its operation, modification and maintenance. However, to
 * facilitate maintenance we ask that before distributing modified
 * versions of this software, you first contact the authors at
 * oof_manager@ctcms.nist.gov. 
 */

#include <oofconfig.h>
#include <iostream>
#include <string>

#include "engine/SparseLib++/gmres.h"
#include "engine/SparseLib++/bicg.h"
#include "engine/SparseLib++/ilupre_double.h"
#include "engine/csrmatrix.h"
#include "engine/element.h"
#include "engine/equation.h"
#include "engine/flux.h"
#include "engine/field.h"
#include "engine/material.h"
#include "engine/matvec.h"
#include "engine/femesh.h"
#include "engine/nodalequation.h"
#include "engine/node.h"
#include "engine/ooferror.h"
#include "engine/preconditioner.h"
#include "engine/solver.h"
#include "engine/stiffnessmatrix.h"
#include "common/tostring.h"
#include "common/trace.h"
#include "common/printvec.h"
#include "common/switchboard.h"
#include "common/IO/progressbar.h"

using namespace std;

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

FEMesh::FEMesh()
  : stiffness(0),
    body_force_rhs(0),
    ncount(0),
    n_defined_dof(0),
    n_active_eqn(0),
    boundary_right_side(0),
    dof_list_needs_cleaning(false),
    nodaleqn_list_needs_cleaning(false),
    precomputeRequired(0),
    microstructure(0),
    forward_mapping(0),
    local_rows_mapping(0)
{
  //  Trace("FEMesh::FEMesh this=" + to_string(this));
  rwlock = new RWLock();
}

FEMesh::FEMesh(CMicrostructure * mc)
  : stiffness(0),
    body_force_rhs(0),
    ncount(0),
    n_defined_dof(0),
    n_active_eqn(0),
    boundary_right_side(0),
    dof_list_needs_cleaning(false),
    nodaleqn_list_needs_cleaning(false),
    precomputeRequired(0),
    microstructure(mc),
    forward_mapping(0),
    local_rows_mapping(0)
{
  //  Trace("FEMesh::FEMesh this=" + to_string(this));
  rwlock = new RWLock();
}

FEMesh::~FEMesh() {
  for(int i=0; i<nodaleqn.size(); i++)
    delete nodaleqn[i];
  for(int i=0; i<dof.size(); i++)
    delete dof[i];
  if(stiffness)
    delete stiffness;
  if(boundary_right_side)
    delete boundary_right_side;
  delete rwlock;
}

const std::string &FEMesh::classname() const {
  static const std::string _name = "FEMesh";
  return _name;
}

const std::string &FEMesh::modulename() const {
  static const std::string _name = "SWIG.engine.femesh";
  return _name;
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

// All Degrees of Freedom are created and deleted with these functions

DegreeOfFreedom *FEMesh::CreateDoF(double x) {
  // localindex is the DoF's position in the node's list of DoFs.
  int index = dof.size();
  DegreeOfFreedom *newdof = new DegreeOfFreedom(x, index);
  dof.push_back(newdof);
  return newdof;
}

void FEMesh::RemoveDoF(DegreeOfFreedom *deaddof) {
//   Trace("FEMesh::RemoveDoF");
  int index = deaddof->dofindex();
  delete deaddof;
  // Don't shift the rest of the vector entries to fill the hole
  // caused by deleting this dof.  The hole can be filled more
  // efficiently by garbage collection later.  Just make sure to check
  // for null pointers when using the dof list!
  dof[index] = 0;
  dof_list_needs_cleaning = true;
}

void FEMesh::clean_doflist() {
  if(dof_list_needs_cleaning) {
    int j=0;
    for(int i=0; i<dof.size(); i++) {
      if (dof[i] != 0) {
	if(i != j) {
	  dof[j] = dof[i];
	  dof[j]->index_ = j;
	}
	j++;
      }
    }
    if(j != dof.size())
      dof.resize(j);
    dof_list_needs_cleaning=false;
  }
}

// When a Node is created *after* the Fields have been defined, the
// degrees of freedom in the Node need to be activated.  Nodes created
// before the Fields are defined have their degrees of freedom
// activated when the Fields are activated.

void FEMesh::activate_active_dofs(FuncNode *n) const {
  for(int i=0; i<Field::allfields().size(); i++) {
    Field *field = Field::allfields()[i];
    if(field->is_active(this)) {
      for(int k=0; k<field->ndof(); k++) {
	(*field)(this, n, k)->activate();
      }
    }
  }
}

int FEMesh::n_defined_dofs() const {
  return n_defined_dof;
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

// All NodalEquations are created and deleted with these functions

NodalEquation *FEMesh::CreateNodalEqn() {
//   Trace("FEMesh::CreateNodalEqn");
  int index = nodaleqn.size();
  NodalEquation *neweqn = new NodalEquation(index);
  nodaleqn.push_back(neweqn);
  return neweqn;
}

void FEMesh::RemoveNodalEqn(NodalEquation *eqn) {
//   Trace("FEMesh::RemoveNodalEqn");
  int index = eqn->index();
  delete eqn;
  // Don't shift the rest of the vector entries to fill the hole
  // caused by deleting this eqn.  The hole can be filled more
  // efficiently by garbage collection later.  Just make sure to check
  // for null pointers when using the dof list!
  nodaleqn[index] = 0;
  nodaleqn_list_needs_cleaning=true;
}

void FEMesh::clean_nodaleqn() {
  if(nodaleqn_list_needs_cleaning) {
    int j = 0;
    for(int i=0; i<nodaleqn.size(); i++) {
      if(nodaleqn[i] != 0) {
	if(i != j) {
	  nodaleqn[j] = nodaleqn[i];
	  nodaleqn[j]->index_ = j;
	}
	j++;
      }
    }
    if(j != nodaleqn.size())
      nodaleqn.resize(j);
    nodaleqn_list_needs_cleaning = false;
  }
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

// Make sure that all dofs and nodalequations have the right indices

void FEMesh::housekeeping() {
  clean_doflist();
  clean_nodaleqn();
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

Node *FEMesh::newMapNode(const Coord &pos) {
  Node *node = new Node(ncount++, pos);
  mapnode.push_back(node);
  return node;
}

FuncNode *FEMesh::newFuncNode(const Coord &pos) {
  FuncNode *node = new FuncNode(this, ncount++, pos);
  funcnode.push_back(node);
  return node;
}

int FEMesh::nnodes() const {
  // don't just return ncount here, since the node index is just an
  // identifier.  When nodes are deleted, ncount won't be decreased.
  return funcnode.size() + mapnode.size();
}

void FEMesh::AddElement(Element *el) {
  element.push_back(el);
  el->set_index(element.size()-1);
}

Element *FEMesh::getElement(int i) const {
  return element[i];
}

int FEMesh::nelements() const {
  return element.size();
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

FEMesh::EquationData::EquationData()
  : active(0),
    active_index(-1)
{}

void FEMesh::activate_equation(Equation &eqn) {
  // Trace("FEMesh::activate_equation " + eqn.name());
  int index = eqn.index();
  // Make sure that there's an EquationData entry for this Equation
  if(index >= eqndata.size()) {
    eqndata.resize(index+1);
  }
  if(!eqndata[index].active) {
    // Mark the Equation as active
    eqndata[index].active = true;
    // Offset of this Equation in the Nodes' NodalEquation lists
    eqndata[index].active_index = n_active_eqn;
    // Total number of active NodalEquations at each Node.
    n_active_eqn += eqn.dim();
    // Create NodalEquations for this Equation in each Node
    for(FuncNodeIterator ni(this); !ni.end(); ++ni) {
      (*this)[ni]->AddEquation(this, eqn);
    }
    // Activate fluxes required for this Equation
    eqn.activate_fluxes(this);
  }
}


void FEMesh::deactivate_equation(Equation &eqn) {
  if(is_active_equation(eqn)) {
    int index = eqn.index();
    eqndata[index].active = false;
    for(FuncNodeIterator ni(this); !ni.end(); ++ni) {
      (*this)[ni]->RemoveEquation(this, eqn);
    }
    // perform garbage collection on the nodaleqn list
    clean_nodaleqn();

    // update each active Equation's active_index
    n_active_eqn = 0;
    for(int i=0; i<eqndata.size(); i++) {
      if(eqndata[i].active) {
	eqndata[i].active_index = n_active_eqn;
	n_active_eqn += Equation::alleqns()[i]->dim();
      }
    }
    eqn.deactivate_fluxes(this);
  }
}

int FEMesh::active_equation_index(const Equation &eqn) const {
  // Offset of this Equation in the Nodes' NodalEquation lists
  return eqndata[eqn.index()].active_index;
}

bool FEMesh::is_active_equation(const Equation &eqn) const {
  int index = eqn.index();
  return (index < eqndata.size()) && eqndata[index].active;
}

int FEMesh::n_active_eqns() const {
  return n_active_eqn;
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

void FEMesh::activate_flux(const Flux &flux) {
  //  std::cerr << "FEMesh::activate_flux: flux = " << flux.name() << std::endl;
  int index = flux.index();
  if(index >= active_flux.size()) {
    active_flux.resize(index+1, 0);
  }
  ++active_flux[index];
}

void FEMesh::deactivate_flux(const Flux &flux) {
  int index = flux.index();
  if(index < active_flux.size()) {
    if(active_flux[index] > 0)
      --active_flux[index];
  }
}

bool FEMesh::is_active_flux(const Flux &flux) const {
  int index = flux.index();
  if(index < active_flux.size()) {
    return active_flux[index] > 0;
  }
  return false;
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

// activate, etc, for Fields have to be routed through Field virtual
// functions, because they have to know if the Field is a
// CompoundField or not.  The routines that do the real work are
// FEMesh::do_activate, etc.

 class FieldMsg : public OOFMessage {
 public:
   FieldMsg(const std::string &msg, const FEMesh &mesh, const Field &field)
     : OOFMessage(msg)
   {
   //     Trace("FieldMsg: " + field.name() + " " + field.classname() + " " + to_string(&field));
     addarg(mesh);
     addarg(field.name());
   }
 };

Field &FEMesh::define_field(Field &field) {
  // Trace("FEMesh::define_field " + field.name());
  field.define(this);
  return field;
}

void FEMesh::undefine_field(Field &field) {
//   Trace("FEMesh::undefine_field " + field.name());
  field.undefine(this);
}

void FEMesh::activate_field(Field &field) {
  // Trace("FEMesh::activate_field " + field.name());
  field.activate(this);
}

void FEMesh::deactivate_field(Field &field) {
//   Trace("FEMesh::deactivate_field " + field.name());
  field.deactivate(this);
}

// Function to copy a field from another mesh.  Other mesh must have
// the same geometry, i.e. the same-indexed funcnodes must be in the
// same location, and the field must be defined.  Called by
// meshmenu.copyMesh.
void FEMesh::acquire_field_data(Field &field, const FEMesh *other) {
  for(FuncNodeIterator i(this); !i.end(); ++i) {
    for(int d=0; d<field.ndof(); ++d) {
      field(this, (*this)[i], d)->value() = \
	field(other, (*other)[i], d)->value();
    }
  }
}

FEMesh::FieldData::FieldData()
  : defined_index(-1),
    active(0),
    defined(0),
    in_plane(0)
{}

void FEMesh::do_define_field(Field &field) {
  // Trace("FEMesh::do_define_field " + field.name() + " " + to_string(field.index()));
  requirePrecompute();
  int index = field.index();
  if(fielddata.size() <= index) {
    fielddata.resize(index+1);
  }
  FieldData &fdata = fielddata[index];
  if(!fdata.defined) {
    fdata.defined = true;
    fdata.defined_index = n_defined_dof;
    n_defined_dof += field.ndof();
    for(FuncNodeIterator ni(this); !ni.end(); ++ni) {
      (*this)[ni]->AddField(this, field);
    }
  }
}

int FEMesh::defined_index(const Field &field) const {
  return fielddata[field.index()].defined_index;
}

bool FEMesh::is_defined(const Field &field) const {
  if(field.index() < fielddata.size())
    return fielddata[field.index()].defined;
  return false;
}

void FEMesh::do_undefine_field(Field &field) {
//   Trace("FEMesh::do_undefine_field " + field.name())
  requirePrecompute();
  if(is_defined_field(field)) {
    if(is_active_field(field)) {
      deactivate_field(field);
    }
    int index = field.index();
    if(index < fielddata.size()) {
      fielddata[index].defined = false;
      for(FuncNodeIterator ni(this); !ni.end(); ++ni) {
	(*this)[ni]->RemoveField(this, field);
      }
      clean_doflist();

      // update each defined Field's defined_index
      n_defined_dof = 0;
      for(int i=0; i<fielddata.size(); i++) {
	if(fielddata[i].defined) {
	  fielddata[i].defined_index = n_defined_dof;
	  n_defined_dof += Field::allfields()[i]->ndof();
	}
      }
    }
  }
}

bool FEMesh::is_defined_field(const Field &field) const {
  return (field.index() < fielddata.size()) && fielddata[field.index()].defined;
}

void FEMesh::do_activate_field(Field &field) {

  if(!is_defined_field(field)) {
    define_field(field);	// allocates space in fielddata array
  }
  if(!is_active_field(field)) {
//     if(field.index() >= fielddata.size()) {
//       fielddata.resize(field.index()+1);
//     }
    FieldData &fdata = fielddata[field.index()];
    fdata.active = true;
//     fdata.active_index = n_active_dof;
//     n_active_dof += field.ndof();
    // Set flags in degrees of freedom
    for(FuncNodeIterator ni(this); !ni.end(); ++ni) {
      for(int i=0; i<field.ndof(); i++) {
	field(this, (*this)[ni], i)->activate();
      }
    }
  }
}

void FEMesh::do_deactivate_field(Field &field) {
  if(is_active_field(field)) {
    fielddata[field.index()].active = false;
//    // update each active Field's active_index
//     n_active_dof = 0;
//     for(int i=0; i<fielddata.size(); i++) {
//       if(fielddata[i].active) {
// 	fielddata[i].active_index = n_active_dof;
// 	n_active_dof += Field::allfields()[i]->ndof();
//       }
//     }
    // Unset flags in degrees of freedom
    for(FuncNodeIterator ni(this); !ni.end(); ++ni) {
      for(int i=0; i<field.ndof(); i++) {
	field(this, (*this)[ni], i)->deactivate();
      }
    }
  }
}

bool FEMesh::is_active_field(const Field &field) const {
  return (field.index() < fielddata.size()) && fielddata[field.index()].active;
}

void FEMesh::set_in_plane(const Field &field, bool planarity) {
  if(field.index() >= fielddata.size()) {
    fielddata.resize(field.index()+1);
  }
  fielddata[field.index()].in_plane = planarity;
}

bool FEMesh::in_plane(const Field &field) const {
  return (field.index() < fielddata.size())
    && fielddata[field.index()].in_plane;
}


//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

// Return a list of all the defined compound fields.  Compound fields
// are the ones that correspond directly to physical fields.
std::vector<CompoundField*>* FEMesh::all_fields() const {
  std::vector<CompoundField*>* flist = new std::vector<CompoundField*>;
  std::vector<CompoundField*> &allfields = CompoundField::allcompoundfields();
  flist->reserve(fielddata.size());
  for(int i=0; i<allfields.size(); i++) {
    if(is_defined_field(*allfields[i])) 
      flist->push_back(allfields[i]);
  }
  return flist;
}

// Likewise for Equation objects.
std::vector<Equation*>* FEMesh::all_equations() const {
  std::vector<Equation*> *eqlist = new std::vector<Equation*>;
  std::vector<Equation*> &alleqns = Equation::alleqns();
  eqlist->reserve(eqndata.size());
  for(int i=0; i<alleqns.size(); i++) {
    if(is_active_equation(*alleqns[i]))
      eqlist->push_back(alleqns[i]);
  }
  return eqlist;
}

std::vector<Flux*>* FEMesh::all_fluxes() const {
  std::vector<Flux*> *fluxlist = new std::vector<Flux*>;
  std::vector<Flux*> &allfluxen = Flux::allfluxes();
  for(int i=0; i<allfluxen.size(); i++) {
    if(is_active_flux(*allfluxen[i]))
      fluxlist->push_back(allfluxen[i]);
  }
  return fluxlist;
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

// void FEMesh::precompute() {
//   // Things that have to be done before anything else can be calculated.
//   Material::precompute_all();
// }

void FEMesh::make_stiffness() {
//   Trace("FEMesh::make_stiffness");
  Progress * localbar = getProgress();
  housekeeping();
  if(stiffness) delete stiffness;
  // Computational part of the right-hand-side.
  if(body_force_rhs) delete body_force_rhs;
 
  stiffness = new StiffnessMatrix(this, nodaleqn.size(), dof.size());
  body_force_rhs = new VECTOR_D(nodaleqn.size(), 0.0);
  
//   precompute();

  Trace("FEMesh::make_stiffness.");
  std::string out_mssg;
  for(int i=0; i<element.size() && !localbar->query_stop(); i++) {
    element[i]->make_stiffness(this, *stiffness, *body_force_rhs);
    localbar->set_progress(float(i+1)/float(element.size()));
    out_mssg = "making stiffness matrix: " + to_string(i+1) + "/" + to_string(element.size());
    localbar->set_message(out_mssg);
  }
}


// Preliminary -- no housekeeping, just run the element's
// post-equilibration loop.
void FEMesh::post_process() {
  //Thread control starts here
  std::string out_mssg;
  Progress * localbar = getProgress(); //not a law firm
  //Thread control ends here
  for(int i=0; i< element.size() && !localbar->query_stop(); i++) {
    element[i]->post_process(this);
    localbar->set_progress(float(i+1)/float(element.size()));
    out_mssg = "postprocessing: " + to_string(i+1) + "/" + to_string(element.size());
    localbar->set_message(out_mssg);
  }
  /*
  if (localbar->get_success() > 0)
    out_mssg = "Succeeded";
  else
    out_mssg = "Failed";
  localbar->set_message(out_mssg);
  */
}



const StiffnessMatrix &FEMesh::stiffness_matrix() const {
  if(!stiffness) {
    throw ErrProgrammingError("Stiffness matrix not computed!",
			      __FILE__, __LINE__);
  } 
  return *stiffness;
}

// Analogous to the above.
VECTOR_D &FEMesh::boundary_rhs() const {
  if(!boundary_right_side) {
    throw ErrProgrammingError("Boundary right hand side not available!",
			      __FILE__, __LINE__);
  }
  return *boundary_right_side;
}

void FEMesh::boundary_rhs_init(int n) {
  if (boundary_right_side) {
    delete boundary_right_side;
    // *boundary_right_side = 0.0; // Assigns to all of the elements.
  }
  boundary_right_side = new VECTOR_D(n, 0.0);
}

void FEMesh::dump_boundary_rhs() {
  std::cerr << "Boundary RHS:" << std::endl;
  for(int i=0; i<boundary_right_side->size(); i++) {
    std::cerr << i << " : " << (*boundary_right_side)[i] << std::endl;
  }
}

void FEMesh::refreshMaterials(PyObject *skeleton) {
  requirePrecompute();
  for(ElementIterator ei=element_iterator(); !ei.end(); ++ei)
    ei.element()->refreshMaterial(skeleton);
}

void FEMesh::requirePrecompute() {
  precomputeLock.acquire();
  precomputeRequired = true;
  precomputeLock.release();
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

NodeIterator::NodeIterator(const FEMesh *const m, NodeIteratorType t)
  : mesh(m),
    type(t),
    index(0)
{}

NodeIterator FEMesh::node_iterator(NodeIteratorType t) const {
  return NodeIterator(this, t);
}

void NodeIterator::operator++() {
  if(index != mesh->nnodes())
    index++;
}

bool NodeIterator::begin() const {
  return index == 0;
}

bool NodeIterator::end() const {
  return index == mesh->nnodes();
}

// Caution: NodeIterator::index is not necessarily the same as
// node.index().  They are only the same if FEMesh::renumber_nodes() has
// been called.  The argument to FEMesh::getNode is the
// NodeIterator::index.

Node *FEMesh::operator[](const NodeIterator &ni) const {
  return getNode(ni.index);
}

Node *NodeIterator::node() const {
  return mesh->getNode(index);
}

void FEMesh::renumber_nodes() {	// Is this ever used?
  int count = 0;
  for(NodeIterator ni = node_iterator(); !ni.end(); ++ni) {
    ni.node()->set_index(count++);
  }
}

void FEMesh::renumber_elements() { // Is this ever used?
  int count = 0;
  for(ElementIterator ei = element_iterator(); !ei.end(); ++ei) {
    ei.element()->set_index(count++);
  }
}

Node *FEMesh::getNode(int i) const {
  if(i < funcnode.size())
    return funcnode[i];
  return mapnode[i - funcnode.size()];
}

// Finding the closest node to the mouse point.
// Used in MeshInfo
// TODO LATER: use hash table lookup here, instead of looping over all nodes.
Node *FEMesh::closestNode(const double x, const double y) {
  double min, dist;
  Node *node, *thenode;
  for(NodeIterator ni = node_iterator(); !ni.end(); ++ni) {
    node = ni.node();
    if (ni.begin()) {
      min = (node->position()(0) - x)*(node->position()(0) - x)+
	    (node->position()(1) - y)*(node->position()(1) - y);
      thenode = node;
    }
    else {
      dist = (node->position()(0) - x)*(node->position()(0) - x)+
	     (node->position()(1) - y)*(node->position()(1) - y);
      if (dist <= min) {
	min = dist;
	thenode = node;
      }
    }
  }
  return thenode;
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//


FuncNodeIterator::FuncNodeIterator(const FEMesh * const g, NodeIteratorType t)
  : mesh(g),
    type(t),
    index(0)
{}

FuncNodeIterator FEMesh::funcnode_iterator(NodeIteratorType t) const {
  return FuncNodeIterator(this, t);
}

void FuncNodeIterator::operator++() {
  if(index != mesh->funcnode.size()) index++;
}

bool FuncNodeIterator::end() const {
  return index == mesh->funcnode.size();
}

FuncNode *FEMesh::operator[](const FuncNodeIterator &ni) const {
  return funcnode[ni.index];
}

FuncNode *FuncNodeIterator::node() const {
  return mesh->funcnode[index];
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

ElementIterator::ElementIterator(const FEMesh * const msh, ElementIteratorType t)
  : mesh(msh),
    type(t),
    index(0)
{}

ElementIterator FEMesh::element_iterator(ElementIteratorType t) const {
  return ElementIterator(this, t);
}

void ElementIterator::operator++() {
  if(!end()) index++;
}

bool ElementIterator::end() const {
  return index == mesh->element.size();
}

Element *ElementIterator::element() const {
  return mesh->element[index];
}

int ElementIterator::size() const {
  return mesh->nelements();
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

// Routines used to pick out rows and columns of the stiffness matrix
// when extracting submatrices.

bool FEMesh::solve_for_this_dof(int i) const {
  //dof list corresponds to unsymmetric system
  return dof[i]->isfree() && dof[i]->active();
}

bool FEMesh::dont_solve_for_this_dof(int i) const {
  //dof list corresponds to unsymmetric system
  return !dof[i]->isfree() || !dof[i]->active();
}

bool FEMesh::solve_this_nodal_eqn(int j) const {
  //nodaleqn corresponds to unsymmetric system
  NodalEquation *ndq = nodaleqn[j];
  bool result = !ndq->is_dependent();
  return result;
  // return !nodaleqn[j]->is_dependent();
}

bool FEMesh::dont_solve_this_nodal_eqn(int j) const {
  return nodaleqn[j]->is_dependent();
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

// Extract the submatrices of the stiffness matrix
// The conceptual subdivision is this:
//
//                 free      fixed
//                          |
//   independent      A     | C
//                          |
//                ----------+---     --- 
//     dependent      C'    | B
// 


// New version.  May be broken up into steps later, but for
// now duplicates the old functionality with the new calling
// semantics.  
//

// Rowmapper and Colmapper.
// Build a map of the rows i of the stiffness matrix for
// which f(i) is true, and return it.  Call with f(i)
// discriminating between independent (known-force) vs.
// dependent (unknown-force) equations for rows, or
// with f(i) discriminating between fixed and free
// DOFs for columns.
//
// Do this by manipulating the pre-existing mapping list.
void FEMesh::rcmapper(bool (FEMesh::*f)(int) const,
		    vector<int>& rcmap, int len) const {
  //  Trace("FEMesh::rcmapper");
  //  vector<int> rowmap( nrows(), -1);
  // Double-check the size?
  int nrc = 0; // # of rows or columns in the mapped submatrix
  rcmap.resize(len); 
  for(int i=0; i<len; i++) {
    if ((this->*f)(i)) {
      rcmap[i] = nrc++;
    }
    else {
      rcmap[i] = -1; // -1 is default for rows/columns which don't occur.
    }
  }
}

// Run the maps, creating the lists used to partition the stiffness
// matrix.

void FEMesh::mapdofeqs() {
  Trace("FEMesh::mapdofeqs");
  
  rcmapper(&FEMesh::solve_this_nodal_eqn, indepeqnmap, stiffness->nrows());
  // cerr << "  indepeqnmap= "; print(indepeqnmap, cerr); cerr << endl;
  rcmapper(&FEMesh::dont_solve_this_nodal_eqn, depeqnmap, stiffness->nrows());
  // cerr << "    depeqnmap= "; print(depeqnmap, cerr); cerr << endl;
  rcmapper(&FEMesh::solve_for_this_dof, freedofmap, stiffness->ncols()); 
  // cerr << "   freedofmap= "; print(freedofmap, cerr); cerr << endl;
  rcmapper(&FEMesh::dont_solve_for_this_dof, fixeddofmap, stiffness->ncols()); 
  // cerr << "  fixeddofmap= "; print(fixeddofmap, cerr); cerr << endl;

} 


// Undo BC invocation by resetting dofs and nodalequations to be free
// and independent.  Does not modify "active" flags.
void FEMesh::resetdofeqs() {
  for(int i=0;i<dof.size();i++) 
    dof[i]->free();
  for(int i=0;i<nodaleqn.size();i++) {
    nodaleqn[i]->independent();
    nodaleqn[i]->value() = 0.0;
  }
}


// Former routine from the FloatBC class, makes more sense here.
// This is a private routine that just does one map, called
// from FEMesh::cleanlists(). 
void FEMesh::cleanmap(vector<int>& rcmap) {
  vector<int> histogram(rcmap.size(),0);
  vector<int> subtractor(rcmap.size(),0);

  // Make the histogram.
  for(int i=0;i<rcmap.size();i++) {
    if (rcmap[i]!=-1)
      histogram[ rcmap[i] ]++;
  }
  // To actually clean the list, the technique is this:
  // Move along the histogram, counting the number of zero
  // entries you see.  When you get to a nonzero entry, 
  // make an entry in the subtraction map -- the amount
  // to subtract from the value you're looking at is the
  // current zero count.  Then, make a pass through the
  // subtraction map and do the actual subtractions.
  int zcount = 0;
  for(int i=0 ; i<rcmap.size(); i++) {
    if(histogram[i]==0) {
      zcount++;
    }
    else {
      subtractor[i] = zcount;
    }
  }

  // Now do the subtraction.
  for(int i=0; i<rcmap.size(); i++) {
    if( rcmap[i] != -1 ) {
      rcmap[i] -= subtractor[ rcmap[i] ];
    }
  }
  

}


// The cleanlists operation restores the row and column maps to 
// consistency after the application of a FloatBC.  The FloatBC's
// cause several rows or columns of the original stiffness matrix
// to be mapped (summed, actually) to a single row or column of
// the extracted 'A' matrix used in the actual linear-algebra solver.
// As a result, the range of the maps can develop holes -- cleanlists()
// closes those holes, ensuring that the range of the map is a 
// contiguous group of integers from 0 to the appropriate N.
void FEMesh::cleanlists() {
  cleanmap(indepeqnmap);
  cleanmap(depeqnmap);
  cleanmap(freedofmap);
  cleanmap(fixeddofmap);
}


//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

// Extract the submatrices of the stiffness matrix
// Mapping arrays should be set up by the time you call this.

CSRmatrix FEMesh::Amatrix() const {
//   Trace("FEMesh::Amatrix");
  return stiffness->submatrix(indepeqnmap,freedofmap);
}

CSRmatrix FEMesh::Bmatrix() const {
//   Trace("FEMesh::Bmatrix");
  CSRmatrix B = stiffness->submatrix(depeqnmap,fixeddofmap);
  return B;
}

CSRmatrix FEMesh::Cmatrix() const {
//   Trace("FEMesh::Cmatrix");
  return stiffness->submatrix(indepeqnmap,fixeddofmap);
}

CSRmatrix FEMesh::Cpmatrix() const {
//   Trace("FEMesh::Cpmatrix");
  return stiffness->submatrix(depeqnmap,freedofmap);
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

void FEMesh::get_unknowns(VECTOR_D &x, const vector<int> &map) const {
  for(int i=0; i<dof.size(); i++) {
    if(map[i] != -1) {
      x[map[i]] += dof[i]->value();
    }
  }
}

void FEMesh::set_unknowns(const VECTOR_D &x, const vector<int> &map) {
//   Trace("FEMesh::set_unknowns");
  for(int i=0; i<dof.size(); i++) {
    if(map[i] != -1) {
      dof[i]->value() = x[map[i]];
    }
  }
}

void FEMesh::get_rhs(VECTOR_D &rhs, const vector<int> &map) const {
//   Trace("FEMesh::get_rhs");
  for(int i=0; i<nodaleqn.size(); i++) {
    if(map[i] != -1) {
      // rhs[map[i]] += nodaleqn[i]->value() + (*boundary_right_side)[i];
      rhs[map[i]] += (*body_force_rhs)[i] + (*boundary_right_side)[i];
      // Merge nodal equation values with boundary RHS contribs.
    }
  }
}

// Writes the rhs data into the nodal equations -- this is not
// really the opposite of get_rhs, beware confusing nomenclature.
void FEMesh::set_rhs(const VECTOR_D &rhs, const vector<int> &map) {
//   Trace("FEMesh::set_rhs");
  for(int i=0; i<nodaleqn.size(); i++) {
    if(map[i] != -1) {
      nodaleqn[(*forward_mapping)[i]]->value() = rhs[map[i]];
    }
  }
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

// Compute the values of the free degrees of freedom
//  A*x = f - C*xp

bool FEMesh::solve(LinearSolver * solver) {

  //at this point the solver and the preconditioner are already selected
  // at the UI and passed as arguments into the solver

  //   Trace("FEMesh::solve:");
  
  //   std::cerr << "Free DOF map." << std::endl;
  //   for(int i=0; i<freedofmap.size(); i++) {
  //     std::cerr << i << " : " << freedofmap[i] << std::endl;
  //   }
  
  //   std::cerr << "Fixed DOF map." << std::endl;
  //   for(int i=0; i<fixeddofmap.size(); i++) {
  //     std::cerr << i << " : " << fixeddofmap[i] << std::endl;
  //   } 
  
  //   std::cerr << "Independent equation map." << std::endl;
  //   for(int i=0; i<indepeqnmap.size(); i++) {
  //     std::cerr << i << " : " << indepeqnmap[i] << std::endl;
  //   }
  
  //   std::cerr << "Dependent equation map." << std::endl;
  //   for(int i=0; i<depeqnmap.size(); i++) {
  //     std::cerr << i << " : " << depeqnmap[i] << std::endl;
  //   }


  // Stiffness output.
  // std::cerr << *stiffness << std::endl;

  CSRmatrix A = Amatrix();

  
  if(A.nrows() != A.ncols())
    throw ErrSetupError("Stiffness matrix is not square! nrows="
			+ to_string(A.nrows()) + ", ncols="
			+ to_string(A.ncols()));
  CSRmatrix C = Cmatrix();

  int n = A.nrows();
  int m = C.ncols();
  VECTOR_D x(n, 0.0);
  VECTOR_D rhs(n, 0.0);
  VECTOR_D xp(m, 0.0);

  rwlock->read_acquire();
  get_unknowns(x, freedofmap);

  get_rhs(rhs, indepeqnmap);
  get_unknowns(xp, fixeddofmap);
  rwlock->read_release();

  rhs -= C*xp;
 
//   std::cerr << "A: " << std::endl;
//   std::cerr << A << std::endl;

//   std::cerr << "Rhs: " << std::endl;
//   for(int i=0; i<rhs.size(); i++) {
//     std::cerr << i << "  " << rhs[i] << std::endl;
//   }
  
  // A *= -1.0;  // Make the matrix +ve-definite for preconditioners.
  //rhs *= -1.0;
  
  //TO THIS POINT, ALL SOLVERS SHOULD DO THE SAME
  

  //THIS SECTION IS FOR DEBUGGING
//   int n_debug = 5;
//   SparseLinkMatrix<double> A_slm(n_debug,n_debug);
//   A_slm(0,0) = 5.0;
//   A_slm(1,1) = 5.0;
//   A_slm(1,0) = 1.0;
//   A_slm(0,1) = 1.0;
//   A_slm(2,2) = 10.0e10;
//   A_slm(3,3) = 10.0e10;
//   A_slm(4,4) = 10.0e10;
  
//   SPARSE_MATRIX_D A_temp(A_slm);
//   //std::cerr << "A_temp at " << &A_temp << std::endl;
//   CSRmatrix A_debug (A_temp);
  
//   VECTOR_D rhs_debug(n_debug, 1.0);
//   VECTOR_D x_debug(n_debug, 0.0);
//   //Assume solver and preconditioner have been already selected
//   //Assume parameters have been set
//   //cerr << "x = \n" << x_debug <<endl;
//   solver->precompute(A_debug);
//   cerr <<"A = \n" << A_debug << endl;
//   solver->solve(A_debug, rhs_debug, x_debug);
//   cerr << "x = \n" << x_debug <<endl;
  
//   solver->postcompute();
  //SECTION IS FOR DEBUGGING ENDS HERE
  solver->set_blocking(stiffness->rowgroups(indepeqnmap),
		       stiffness->colgroups(freedofmap));
  
  solver->precompute(A);
  //cerr <<"A = \n" << A << endl;
  bool status = solver->solve(A, rhs, x);
  //cerr << "x = \n" << x <<endl;
  // std::cerr << "rhs = \n" << rhs << std::endl;
  // std::cerr << solver->message() << std::endl;
  solver->postcompute(); // Call possibly-null post-compute hook.

  rwlock->write_acquire();

//   std::cerr << "Setting unknonwns -- free DOF map." << std::endl;
//   for(int i=0; i<freedofmap.size(); i++) {
//     std::cerr << i << " : " << freedofmap[i] << std::endl;
//   }
  
  set_unknowns(x, freedofmap);
  
  //   std::cerr << "x on exit from the solver." << std::endl;
  //   for(int i=0; i<x.size(); i++) {
  //     std::cerr << i << " : " << x[i] << std::endl;
  //   }
  
  CSRmatrix Cp = Cpmatrix();
  CSRmatrix B = Bmatrix();
  
  VECTOR_D rhsp = Cp*x+B*xp;
  set_rhs(rhsp, depeqnmap); // <-- not modified to reflect local_rhs.
  rwlock->write_release();

  return status;
}






//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

void FEMesh::dump_dof() const {
  for(FuncNodeIterator i(this); !i.end(); ++i) {
    const FuncNode &node = *(*this)[i];
    cout << "node = " << node << endl;
    for(int f=0; f<Field::allfields().size(); f++) {
      const Field &field = *Field::allfields()[f];
      if(field.is_defined(this)) {
	for(int j=0; j<field.ndof(); j++) {
	  cout << "\t" << field.name() << "[" << j << "] dof="
	       << *field(this, node, j) << endl;
	}
      }
    }
    for(int e=0; e<Equation::alleqns().size(); e++) {
      const Equation &eqn = *Equation::alleqns()[e];
      if(eqn.is_active(this)) {
	for(int j=0; j<eqn.dim(); j++) {
	  cout << "\t" << eqn.name() << "[" << j << "] nodaleqn="
	       << *eqn.nodaleqn(this, node, j) << endl;
	}
      }
    }
  }
}


void FEMesh::dump_eqn() const {
  for(FuncNodeIterator i(this); !i.end(); ++i) {
    const FuncNode &node = *(*this)[i];
    cout << "node = " << node << endl;
    for(int e=0; e<Equation::alleqns().size(); e++) {
      const Equation &eqn = *Equation::alleqns()[e];
      if(eqn.is_active(this)) {
	for(int j=0; j<eqn.dim(); j++) {
	  cout << "\t" << eqn.name() << "[" << j << "] nodaleqn="
	       << *eqn.nodaleqn(this, node, j) << endl;
	}
      }
    }
  }
}


void FEMesh::set_local_mapping(const std::vector<int>&rows, 
		       const std::vector<int>&cols)
{
  //allocate memory for local mapping
  //clear_symmetricity_maps();
  if(local_rows_mapping)
    {
      delete local_rows_mapping;  
    }
  local_rows_mapping = new std::vector<int> (rows.size());


  // construct mapping to relate the local nodal equations to the 
  // the to-solve-for degrees of freedom
  for(int irow = 0; irow< rows.size(); irow++)
    {
      (*local_rows_mapping)[rows[irow]] = cols[irow];
      //std::cerr << irow <<"=" << cols[irow] << std::endl; 
    }
  //allocate memory for the forward and backward mapping
  int global_mapping_size = nodaleqn.size(); //number of rows
  if (forward_mapping)
    {
      delete forward_mapping;
      
    }
  forward_mapping = new std::vector<int> (global_mapping_size);

}

void FEMesh::clear_symmetricity_maps()
{
  //destroy the symmetricity maps after using them
  delete forward_mapping;
  delete local_rows_mapping;
}

const NodalEquation & FEMesh::map_forward_local(const Equation & equation, 
			const FuncNode &node, const int &eqcomp)
{
  int local_target_row = (*local_rows_mapping)[equation.localindex(this, eqcomp)];
  return *node.eqnlist[local_target_row];
}


int FEMesh::map_forward(const Equation & equation, 
			const FuncNode &node, const int &eqcomp) const 
{
  /*
    Finds the row on the rhs vector and the global stiffness matrix 
    where to insert/read the FE component, so that the stiffness matrix
    is symmetric, according to a local mapping of the equations.
  */
  int original_row = equation.nodaleqn(this, node, eqcomp)->index();
  
  int local_target_row = (*local_rows_mapping)[equation.localindex(this, eqcomp)];
  const NodalEquation * local_target_equation = node.eqnlist[local_target_row];
  
  int target_row = local_target_equation->index(); 
  (*forward_mapping)[original_row] = target_row;
  return target_row;
}


int FEMesh::check_symmetricity(const CSRmatrix &A)
{
  for (int i=0; i<A.nrows(); i++)
    for (int j=0; j<A.ncols(); j++)
      if (A(i,j) != A(j,i))
	{
	  std::cerr << "(i,j)= (" << i <<"," << j << ")"<< std::endl;
	  std::cerr << "A(i,j)=" << A(i,j) << " A(j,i)=" << A(j,i) << std::endl;
	  return 0;
	}
  return 1;

}
