# -*- python -*-
# $RCSfile: problem.py,v $
# $Revision: 1.33 $
# $Author: redwing $
# $Date: 2005/01/31 23:31:39 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov. 


from SWIG.common import switchboard
from SWIG.common.space import spacedim
from SWIG.engine import conjugate
from SWIG.engine import equation
from SWIG.engine import field
from SWIG.engine import fieldindex
from SWIG.engine import flux
from common import debug
from common import utils
from common.IO import xmlmenudump
import sys
import types


def define(obj):
    utils.OOFdefine(obj.name(), obj)
    return obj

def defineField(fld):
    define(fld)
    out_of_plane = fld.out_of_plane();
    define(out_of_plane)
    # "new field" is sent here, instead of from the Field constructor,
    # because it must be called *after* the field is defined in the
    # OOF namespace.
    switchboard.notify("new field")
    return fld

def defineFlux(flx):
    define(flx)
    switchboard.notify("new flux")
    return flx

def defineEquation(eqn):
    define(eqn)
    switchboard.notify("new equation")
    return eqn

####################

# Define a field.  This creates an object named 'Temperature' in the
# OOF namespace.
Temperature = defineField(field.ScalarField('Temperature'))
# Define a flux
Heat_Flux = defineFlux(flux.VectorFlux('Heat_Flux'))
# And equations
HeatBalanceEquation = defineEquation(equation.DivergenceEquation('Heat_Eqn', Heat_Flux, 1))
HeatOutOfPlane = defineEquation(equation.PlaneFluxEquation('Plane_Heat_Flux', Heat_Flux, 1))

## this creates the Displacement, Stress, and Mechanical Equilibrium equations
Displacement = defineField(field.TwoVectorField('Displacement'))
Stress = defineFlux(flux.SymmetricTensorFlux('Stress'))

ForceBalanceEquation = defineEquation(equation.DivergenceEquation('Force_Balance', Stress, spacedim))
ForcesOutOfPlane = defineEquation(equation.PlaneFluxEquation('Plane_Stress', Stress, 3))


## Define electrostatic potential
Voltage = defineField(field.ScalarField('Voltage'))
## Define total polarization vector
Total_Polarization = defineFlux(flux.VectorFlux('Total_Polarization'))
## Differential form of Coulomb's Law
CoulombEquation = defineEquation(equation.DivergenceEquation('Coulomb_Eqn', Total_Polarization, 1))
PolarizationOutOfPlane = defineEquation(equation.PlaneFluxEquation('InPlanePolarization', Total_Polarization, 1))
## Define conjugate quantities for all fields and equations
##
##
## Force balance equation
##
## In-plane components (u,v)
##
## The 3D displacement vector is (u, v, w),
## where
## u lies along the x-direction,
## v, along the y-direction,
## and w, out-of-the-plane, or along
## the z-direction
u = fieldindex.VectorFieldIndex(0)
v = fieldindex.VectorFieldIndex(1)

fx = fieldindex.VectorFieldIndex(0)
fy = fieldindex.VectorFieldIndex(1)
conjugate.ConjugatePair("Elasticity", ForceBalanceEquation, fx,
                        Displacement, u) ## fx corresponds to u
conjugate.ConjugatePair("Elasticity", ForceBalanceEquation, fy,
                        Displacement, v) ## fy is conjugate to v

## out-of-plane compoments
## The available out-of-plane components of stress are
## $\sigma_{zz}, \sigma_{zy}, \sigma_{xy}$, in *that* order,
## (0, 1, 2).
## the out-of-plane displacement is $\frac{\partial^2 w}{\pratial z^2}$
u_xz = fieldindex.VectorFieldIndex(0)
u_yz = fieldindex.VectorFieldIndex(1)
w_zz = fieldindex.VectorFieldIndex(2)
sigma_xz = fieldindex.OutOfPlaneSymTensorIndex(2,1)
sigma_yz = fieldindex.OutOfPlaneSymTensorIndex(2,0)
sigma_zz = fieldindex.OutOfPlaneSymTensorIndex(2,2)

conjugate.ConjugatePair("Elasticity",ForcesOutOfPlane, sigma_zz, 
                        Displacement.out_of_plane(), w_zz)
 ##  $\sigma_{zz}$ is conjugate to $\frac{\partial^2 w}{ \partial z^2}$

conjugate.ConjugatePair("Elasticity",ForcesOutOfPlane, sigma_xz, 
                        Displacement.out_of_plane(), u_xz)
 ##  $\sigma_{xz}$ is conjugate to $\frac{\partial^2 u}{ \partial x \partial z}$

conjugate.ConjugatePair("Elasticity",ForcesOutOfPlane, sigma_yz, 
                        Displacement.out_of_plane(), u_yz)
 ##  $\sigma_{yz}$ is conjugate to $\frac{\partial^2 u}{ \partial y \partial z}$

###############################################################
##
## Heat flux equation
##
## In-plane components, T
##
T = fieldindex.ScalarFieldIndex()
DivJ = fieldindex.ScalarFieldIndex()

conjugate.ConjugatePair("Thermal:Conductivity", HeatBalanceEquation, DivJ,
                        Temperature, T)
 ## $\nabla \cdot \vec{J}$ is conjugate to T

## out-of-plane components, $\frac{\partial T}{\partial z}$
T_z = fieldindex.OutOfPlaneVectorFieldIndex(2)
J_z = fieldindex.OutOfPlaneVectorFieldIndex(2)

conjugate.ConjugatePair("Thermal:Conductivity", HeatOutOfPlane, J_z,
                        Temperature.out_of_plane(), T_z)
 ##  $J_{z}$ is conjugate to $\frac{\partial T}{\partial z}$

###############################################################
##
## Coulomb equation
##
## In-plane components, phi
##
phi = fieldindex.ScalarFieldIndex()
DivD = fieldindex.ScalarFieldIndex()

conjugate.ConjugatePair("Electric:DielectricPermittivity", CoulombEquation, DivD,
                        Voltage, phi)
 ## $\nabla \cdot \vec{D}$ is conjugate to D

## out-of-plane components, $\frac{\partial D}{\partial z}$
phi_z = fieldindex.OutOfPlaneVectorFieldIndex(2)
D_z = fieldindex.OutOfPlaneVectorFieldIndex(2)

conjugate.ConjugatePair("Electric:DielectricPermittivity", PolarizationOutOfPlane, D_z,
                        Voltage.out_of_plane(), phi_z)
 ##  $D_{z}$ is conjugate to $\frac{\partial phi}{\partial z}$



###########################

def xmldump(file):
    print >> file, "<section id='Section:Builtin'>"
    print >> file, " <title>Built-In Physics</title>"
    print >> file, """<para>
    &oof2; is designed to be extendible, so it is easy to add new
    &fields;, &fluxes;, and &equations;.  That means that the
    following tables of <emphasis>built-in</emphasis> objects may not
    be complete.  The listed dimension is the number of independent
    components in each object.
    </para>"""

    print >> file, "<table>"
    print >> file, " <title>Built-in Fields</title>"
    print >> file, " <tgroup cols='3'>"
    print >> file, "  <thead><row>"
    print >> file, "   <entry>Name</entry>"
    print >> file, "   <entry>Type</entry>"
    print >> file, "   <entry>Dimension</entry>"
    print >> file, "  </row></thead>"
    print >> file, "  <tbody>"
    for fld in field.allFields:
        print >> file, "<row id='Field:%s'>" % fld.name()
        print >> file, " <entry><userinput>%s</userinput></entry>" % fld.name()
        print >> file, " <entry><classname>%s</classname></entry>" \
              % fld.__class__.__name__
        print >> file, " <entry>%d</entry>" % fld.ndof()
        print >> file, "</row>"
    print >> file, "  </tbody>"
    print >> file, " </tgroup>"
    print >> file, "</table>"

    print >> file, "<table>"
    print >> file, " <title>Built-in Fluxes</title>"
    print >> file, " <tgroup cols='3'>"
    print >> file, "  <thead><row>"
    print >> file, "   <entry>Name</entry>"
    print >> file, "   <entry>Type</entry>"
    print >> file, "   <entry>Dimension</entry>"
    print >> file, "  </row></thead>"
    print >> file, "  <tbody>"
    for flx in flux.allFluxes:
        print >> file, "<row id='Flux:%s'>" % flx.name()
        print >> file, " <entry><userinput>%s</userinput></entry>" % flx.name()
        print >> file, " <entry><classname>%s</classname></entry>" \
              % flx.__class__.__name__[:-3] # strip 'Ptr'
        print >> file, " <entry>%d</entry>" % flx.ndof()
        print >> file, "</row>"
    print >> file, "  </tbody>"
    print >> file, " </tgroup>"
    print >> file, "</table>"

    print >> file, "<table>"
    print >> file, " <title>Built-in Equations</title>"
    print >> file, " <tgroup cols='4'>"
    print >> file, "  <thead><row>"
    print >> file, "   <entry>Name</entry>"
    print >> file, "   <entry>Type</entry>"
    print >> file, "   <entry>Flux</entry>"
    print >> file, "   <entry>Dimension</entry>"
    print >> file, "  </row></thead>"
    print >> file, "  <tbody>"
    links = {'PlaneFluxEquation' : 'Section:Concepts:Mesh:Equation:PlaneFlux',
             'DivergenceEquation' : 'Section:Concepts:Mesh:Equation:Divergence'}
    for eqn in equation.allEquations:
        classname = eqn.__class__.__name__[:-3] # strip off 'Ptr'
        print >> file, "<row id='Equation:%s'>" % eqn.name()
        print >> file, " <entry><userinput>%s</userinput></entry>" % eqn.name()
        print >> file, " <entry><link linkend='%s'><classname>%s</classname></link></entry>"  % (links[classname], classname)
        print >> file, " <entry><userinput>%s</userinput></entry>" \
              % eqn.fluxname()
        print >> file, " <entry>%d</entry>" % eqn.ndof()
        print >> file, "</row>"        
    print >> file, "  </tbody>"
    print >> file, " </tgroup>"
    print >> file, "</table>"

    print >> file, "</section>"
    

xmlmenudump.addSection(xmldump)

