// -*- C++ -*-
// $RCSfile: solver.h,v $
// $Revision: 1.31 $
// $Author: redwing $
// $Date: 2005/03/01 16:31:13 $

/* This software was produced by NIST, an agency of the U.S. government,
 * and by statute is not subject to copyright in the United States.
 * Recipients of this software assume all responsibilities associated
 * with its operation, modification and maintenance. However, to
 * facilitate maintenance we ask that before distributing modified
 * versions of this software, you first contact the authors at
 * oof_manager@ctcms.nist.gov. 
 */

#include <oofconfig.h>

#ifndef SOLVER_H
#define SOLVER_H

#include "matvec.h"
#include "preconditioner.h"
#include <string>
class ThreadControl;
class CSRmatrix;
class FEMesh;


class LinearSolver {
public:
  /*
    pointer to correct preconditioner will not be set at
    construction of solver. Will be constructed through a UI
   */
  std::string _message;
  std::vector<int> rowblocks, colblocks;
  LinearSolver() {};
  virtual ~LinearSolver() {}
  virtual void set_blocking(const std::vector<int> rblocks,
			    const std::vector<int> cblocks);
  virtual void set_blocking(const FEMesh *);
  virtual void precompute(const CSRmatrix &A) = 0; 
  virtual void postcompute() = 0;
  virtual bool solve(const CSRmatrix &A, const VECTOR_D &rhs, VECTOR_D &x) = 0;
  virtual bool status() const = 0;
  virtual std::string message() const{ return "void solver" + _message;};
  virtual double get_residual() {return 0;}
};


class SolverMessager {
protected:
  std::string message;
public:
  SolverMessager(const int, const int, const double);
  SolverMessager() {};
  virtual std::string _message() const {return message;};
  virtual ~SolverMessager() {};
};




// Preconditioned gmres solver.

class GMRESSolver : public LinearSolver {
private:
  bool status_;
  int max_iterations;
  double tolerance;
  int number_of_iterations;
  double residual_norm;
  PreconditionerWrap *pc; //wrapper to the actual preconditioner
  int krylov_dimension;
public:  
  GMRESSolver(int max_iterations, int krylov_dimension, double tolerance, PreconditionerWrap * pc)
    : max_iterations(max_iterations), tolerance(tolerance), 
    krylov_dimension(krylov_dimension),
    pc(pc), 
    status_(0),
    number_of_iterations(0),
    residual_norm(0)
    
  {}
  virtual ~GMRESSolver(); 
  virtual double get_residual() {return residual_norm;}
  virtual void precompute(const CSRmatrix &A){};
  virtual void postcompute() {};
  virtual bool solve(const CSRmatrix &A, const VECTOR_D &rhs, VECTOR_D &x);
  virtual bool status() const { return status_; }
  virtual std::string message() const;
};

class BiCGSolver : public LinearSolver {
private:
  bool status_;
  int max_iterations;
  double tolerance;
  int number_of_iterations;
  double residual_norm;
  PreconditionerWrap *pc; //wrapper to the actual preconditioner
public:  
  BiCGSolver(int max_iterations, double tolerance, PreconditionerWrap * pc)
    : max_iterations(max_iterations), tolerance(tolerance),
      pc(pc), 
      status_(0),
    number_of_iterations(0),
    residual_norm(0)
  {}
  ~BiCGSolver();
  virtual double get_residual() {return residual_norm;}
  virtual void precompute(const CSRmatrix &A) {};
  virtual void postcompute() {};
  virtual bool solve(const CSRmatrix &A, const VECTOR_D &rhs, VECTOR_D &x);
  virtual bool status() const { return status_; }
  virtual std::string message() const;
};

class CGSolver : public LinearSolver {
private:
  bool status_;
  int max_iterations;
  double tolerance;
  int number_of_iterations;
  double residual_norm;
  PreconditionerWrap *pc; //wrapper to the actual preconditioner
public:  
  CGSolver(int max_iterations, double tolerance, PreconditionerWrap * pc)
    : max_iterations(max_iterations), tolerance(tolerance),
      pc(pc), 
      status_(0),
    number_of_iterations(0),
    residual_norm(0)
  {}
  ~CGSolver();
  virtual double get_residual() {return residual_norm;}
  virtual void precompute(const CSRmatrix &A) {};
  virtual void postcompute() {};
  virtual bool solve(const CSRmatrix &A, const VECTOR_D &rhs, VECTOR_D &x);
  virtual bool status() const { return status_; }
  virtual std::string message() const;
};

class BiCGStabSolver : public LinearSolver {
private:
  bool status_;
  int max_iterations;
  double tolerance;
  int number_of_iterations;
  double residual_norm;
  PreconditionerWrap *pc; //wrapper to the actual preconditioner
public:  
  BiCGStabSolver(int max_iterations, double tolerance, PreconditionerWrap * pc)
    : max_iterations(max_iterations), 
    tolerance(tolerance),
    pc(pc), 
    status_(0),
    number_of_iterations(0),
    residual_norm(0)
  {}
  ~BiCGStabSolver();
  virtual double get_residual() {return residual_norm;}
  virtual void precompute(const CSRmatrix &A) {};
  virtual void postcompute() {};
  virtual bool solve(const CSRmatrix &A, const VECTOR_D &rhs, VECTOR_D &x);
  virtual bool status() const { return status_; }
  virtual std::string message() const;
};


// For diagnostic purposes, provide a means to solve the Ax=b problem
// directly by a dense-matrix back-substitution method.  This will
// typically be very slow and inefficient, and should not be used for
// real problems of any substantial size.
class DirectSolver : public LinearSolver {
private:
  bool status_;
  double residual_norm;
public:
  DirectSolver() : status_(0), residual_norm(0)
  {}
  ~DirectSolver();
  virtual double get_residual() {return residual_norm;}
  virtual void precompute(const CSRmatrix &A) {};
  virtual void postcompute() {};
  virtual bool solve(const CSRmatrix &A, const VECTOR_D &rhs, VECTOR_D &x);
  virtual bool status() const { return status_; }
  virtual std::string message() const;
};

std::ostream & operator<<(std::ostream &, const LinearSolver &);

//bool linear_solver(Solver *, FEMesh*);

#endif // SOLVER_H








